import os
import subprocess
import sys
from typing import Dict, List, NamedTuple, Union


class QudaEnumMeta(NamedTuple):
    name: str
    value: Union[int, str]

    def __repr__(self) -> str:
        return f"{self.name} = {self.value}"


class QudaParamsMeta(NamedTuple):
    name: str
    type: str
    ptr: str
    array: List[int]

    def __repr__(self) -> str:
        self_array = "" if len(self.array) == 0 else f"[{']['.join(self.array)}]"
        return f"{self.type} {self.ptr}{self.name}{self_array}"


normal = """
    @property
    def %name%(self):
        return self.param.%name%

    @%name%.setter
    def %name%(self, value):
        self.param.%name% = value
"""

cstring = """
    @property
    def %name%(self):
        return self.param.%name%

    @%name%.setter
    def %name%(self, const char value[]):
        self.param.%name% = value
"""

param = """
    @property
    def %name%(self):
        param = %type%()
        param.from_ptr(self.param.%name%)
        return param

    @%name%.setter
    def %name%(self, value):
        self.set_%name%(value)

    cdef set_%name%(self, %type% value):
        self.param.%name% = &value.param
"""

multigrid = """
    @property
    def %name%(self):
        value = []
        for i in range(self.n_level):
            value.append(self.param.%name%[i])
        return value

    @%name%.setter
    def %name%(self, value):
        for i in range(self.n_level):
            self.param.%name%[i] = value[i]
"""

multigrid_param = """
    @property
    def %name%(self):
        params = []
        for i in range(self.param.n_level):
            if self.param.%name%[i] != NULL:
                param = %type%()
                param.from_ptr(self.param.%name%[i])
                params.append(param)
            else:
                params.append(None)
        return params

    @%name%.setter
    def %name%(self, value):
        for i in range(self.param.n_level):
            if value[i] is not None:
                self.set_%name%(value[i], i)
            else:
                self.param.%name%[i] = NULL

    cdef set_%name%(self, %type% value, int i):
        self.param.%name%[i] = &value.param
"""

void_ptr = """
    @property
    def %name%(self):
        ptr = Pointer("void")
        ptr.set_ptr(self.param.%name%)
        return ptr

    @%name%.setter
    def %name%(self, Pointer value):
        assert value.dtype == "void"
        self.param.%name% = value.ptr
"""

ptr = """
    @property
    def %name%(self):
        ptr = Pointer("%type%")
        ptr.set_ptr(<void *>self.param.%name%)
        return ptr

    @%name%.setter
    def %name%(self, Pointer value):
        assert value.dtype == "%type%"
        self.param.%name% = <%type% *>value.ptr
"""

ptrptr = """
    @property
    def %name%(self):
        ptrs = Pointers("%type%", 0)
        ptrs.set_ptrs(<void **>self.param.%name%)
        return ptrs

    @%name%.setter
    def %name%(self, Pointers value):
        assert value.dtype == "%type%"
        self.param.%name% = <%type% **>value.ptrs
"""


def build_pyquda_pyx(pyquda_root, quda_path):
    fake_libc_include = os.path.join(pyquda_root, "pycparser", "utils", "fake_libc_include")
    quda_include = os.path.join(quda_path, "include")
    assert os.path.exists(fake_libc_include), f"{fake_libc_include} not found"
    print(f"Building pyquda wrapper from {os.path.join(quda_include, 'quda.h')}")
    sys.path.insert(1, os.path.join(pyquda_root, "pycparser"))
    try:
        from pycparser import parse_file, c_ast
    except ImportError or ModuleNotFoundError:
        from pycparser.pycparser import parse_file, c_ast  # This is for the language server
    sys.path.remove(os.path.join(pyquda_root, "pycparser"))

    def evaluate(node):
        if node is None:
            return
        elif type(node) is c_ast.UnaryOp:
            if node.op == "+":
                return evaluate(node.expr)
            elif node.op == "-":
                return -evaluate(node.expr)
        elif type(node) is c_ast.BinaryOp:
            if node.op == "+":
                return evaluate(node.left) + evaluate(node.right)
            elif node.op == "-":
                return evaluate(node.left) - evaluate(node.right)
        elif type(node) is c_ast.Constant:
            return int(node.value, 0)
        elif type(node) is c_ast.ID:
            return node.name
        else:
            raise ValueError(f"Unknown node {node}")

    subprocess.run(
        [
            "cpp",
            "-P",
            Rf"-I{quda_include}",
            "-o",
            os.path.join(pyquda_root, "pyquda", "quda_define.py"),
            os.path.join(pyquda_root, "pyquda", "quda_define.in.py"),
        ],
        check=True,
    )

    quda_enum_meta: Dict[str, List[QudaParamsMeta]] = {}
    quda_params_meta: Dict[str, List[QudaParamsMeta]] = {}
    ast = parse_file(
        os.path.join(quda_include, "quda.h"),
        use_cpp=True,
        cpp_path=os.environ["CC"] if "CC" in os.environ else "cc",
        cpp_args=["-E", Rf"-I{fake_libc_include}", Rf"-I{quda_include}"],
    )
    for node in ast:
        if not hasattr(node, "name") or node.name is None:
            continue
        if node.name.startswith("Quda") and type(node.type.type) is c_ast.Enum:
            quda_enum_meta[node.name] = []
            current_value = -1
            for item in node.type.type.values:
                item_value = evaluate(item.value)
                if item_value is not None:
                    current_value = item_value
                else:
                    current_value += 1
                if "INVALID" in item.name:
                    quda_enum_meta[node.name].append(QudaEnumMeta(item.name, "QUDA_INVALID_ENUM"))
                else:
                    quda_enum_meta[node.name].append(QudaEnumMeta(item.name, current_value))
        if node.name.startswith("Quda") and node.name.endswith("Param"):
            quda_params_meta[node.name] = []
            # print(node.name)
            for decl in node.type.type.decls:
                n, t = decl.name, decl.type
                if type(t) is c_ast.Union:
                    n, t = t.decls[0].name, t.decls[0].type
                tt = t.type
                if type(t) is c_ast.TypeDecl:
                    quda_params_meta[node.name].append(QudaParamsMeta(n, " ".join(tt.names), "", []))
                    # print(" ".join(t.type.names), n)
                elif type(t) is c_ast.ArrayDecl:
                    ttt = tt.type
                    if type(tt) is c_ast.TypeDecl:
                        quda_params_meta[node.name].append(QudaParamsMeta(n, " ".join(ttt.names), "", [t.dim.value]))
                        # print(" ".join(t.type.type.names), n, f"[{t.dim.value}]")
                    elif type(tt) is c_ast.PtrDecl:
                        quda_params_meta[node.name].append(
                            QudaParamsMeta(n, " ".join(ttt.type.names), "*", [t.dim.value])
                        )
                        # print(" ".join(t.type.type.type.names), "*", n, f"[{t.dim.value}]")
                    elif type(tt) is c_ast.ArrayDecl:
                        quda_params_meta[node.name].append(
                            QudaParamsMeta(n, " ".join(ttt.type.names), "", [t.dim.value, tt.dim.value])
                        )
                        # print(" ".join(t.type.type.type.names), n, f"[{t.dim.value}][{t.type.dim.value}]")
                    else:
                        raise ValueError(f"Unexpected node {node}")
                elif type(t) is c_ast.PtrDecl:
                    ttt = tt.type
                    if type(tt) is c_ast.TypeDecl:
                        quda_params_meta[node.name].append(QudaParamsMeta(n, " ".join(ttt.names), "*", ""))
                        # print(" ".join(t.type.type.names), "*", n)
                    elif type(tt) is c_ast.PtrDecl:
                        quda_params_meta[node.name].append(QudaParamsMeta(n, " ".join(ttt.type.names), "**", ""))
                        # print(" ".join(t.type.type.type.names), "**", n)
                    else:
                        raise ValueError(f"Unexpected node {node}")
                else:
                    raise ValueError(f"Unexpected node {node}")
        if isinstance(node.type, c_ast.FuncDecl):
            continue
            if node.name in ["set_dim", "pack_ghost"]:
                continue
            p = []
            if node.type.args is not None:
                for arg in node.type.args.params:
                    n, t = arg.name, arg.type
                    tt = t.type
                    if type(t) is c_ast.TypeDecl:
                        p.append((" ".join(t.quals + tt.names + [""]), n))
                    elif type(t) is c_ast.PtrDecl:
                        ttt = tt.type
                        if type(tt) is c_ast.TypeDecl:
                            p.append(
                                (" ".join(tt.quals + ttt.names + ["*"]), " ".join(t.quals + [n]))
                            )  # const type * const
                        elif type(tt) is c_ast.PtrDecl:
                            tttt = ttt.type
                            if type(ttt) is c_ast.TypeDecl:
                                p.append((" ".join(ttt.quals + tttt.names + ["**"]), n))
                            elif type(ttt) is c_ast.PtrDecl:
                                ttttt = tttt.type
                                if type(tttt) is c_ast.TypeDecl:
                                    p.append((" ".join(tttt.quals + ttttt.names + ["***"]), n))
                                else:
                                    raise ValueError(f"Unexpected node {node}")
                            else:
                                raise ValueError(f"Unexpected node {node}")
                        else:
                            raise ValueError(f"Unexpected node {node}")
                    elif type(t) is c_ast.ArrayDecl:
                        ttt = tt.type
                        if t.dim is not None:
                            if type(tt) is c_ast.TypeDecl:
                                p.append((" ".join(tt.quals + ttt.names + [""]), f"{n}[{t.dim.value}]"))
                            elif type(tt) is c_ast.PtrDecl:
                                tttt = ttt.type
                                if type(ttt) is c_ast.TypeDecl:
                                    p.append((" ".join(ttt.quals + tttt.names + ["*"]), f"{n}[{t.dim.value}]"))
                                else:
                                    raise ValueError(f"Unexpected node {node}")
                            else:
                                raise ValueError(f"Unexpected node {node}")
                        else:
                            if type(tt) is c_ast.TypeDecl:
                                p.append((" ".join(tt.quals + ttt.names + [""]), f"{n}[]"))
                            elif type(tt) is c_ast.PtrDecl:
                                tttt = ttt.type
                                if type(ttt) is c_ast.TypeDecl:
                                    p.append((" ".join(ttt.quals + tttt.names + ["*"]), f"{n}[]"))
                                else:
                                    raise ValueError(f"Unexpected node {node}")
                    else:
                        raise ValueError(f"Unexpected node {node}")
            ntt = node.type.type
            p = [""] if p == [("void", None)] else [f"{_t}{_a}" for _t, _a in p]
            if type(ntt) is c_ast.TypeDecl:
                print(" ".join(ntt.quals + ntt.type.names + [""]) + f"{node.name}(" + ", ".join(p) + ")")
            elif type(ntt) is c_ast.PtrDecl:
                nttt = ntt.type
                print(" ".join(nttt.quals + nttt.type.names + ["*"]) + f"{node.name}(" + ", ".join(p) + ")")
            else:
                raise ValueError(f"Unexpected node {node}")

    with open(os.path.join(quda_include, "quda_constants.h"), "r") as f:
        quda_constants_h = f.read()
    with open(os.path.join(quda_include, "enum_quda.h"), "r") as f:
        enum_quda_h = f.read()
    with open(os.path.join(pyquda_root, "pyquda", "enum_quda.in.py"), "r") as f:
        enum_quda_py = f.read()
    with open(os.path.join(pyquda_root, "pyquda", "src", "quda.in.pxd"), "r") as f:
        quda_pxd = f.read()
    with open(os.path.join(pyquda_root, "pyquda", "src", "quda.in.pyx"), "r") as f:
        quda_pyx = f.read()
    # with open(os.path.join(pyquda_root, "pyquda", "quda.in.pyi"), "r") as f:
    #     quda_pyi = f.read()

    quda_constants_py = ""
    quda_constants_pxd = 'cdef extern from "quda_constants.h":\n    cdef enum:'
    start = quda_constants_h.find("#define")
    while start >= 0:
        stop = quda_constants_h.find("\n", start)
        name_value = quda_constants_h[start + len("#define") : stop].strip().split(" ")
        name, value = name_value[0], " ".join(name_value[1:]).strip()
        idx_comment = quda_constants_h.rfind("@brief", 0, start)
        idx_endline = quda_constants_h.find("*/", idx_comment, start)
        if idx_comment < idx_endline:
            comment = "\n".join(
                [
                    line.strip("* ")
                    for line in quda_constants_h[idx_comment + len("@brief") : idx_endline].strip().split("\n")
                ]
            )
            comment = f'"""\n{comment}\n"""\n'
        else:
            comment = ""
        quda_constants_py += f"{name} = {value}\n{comment}\n"
        quda_constants_pxd += f"\n        {name}\n"
        start = quda_constants_h.find("#define", stop)
    enum_quda_py = enum_quda_py.replace("# quda_constants.py\n", quda_constants_py)

    enum_quda_pxd = 'cdef extern from "enum_quda.h":'
    for key, val in quda_enum_meta.items():
        enum_quda_pxd += f"\n    ctypedef enum {key}:\n        pass\n"

    idx_key = 0
    for key, val in quda_enum_meta.items():
        enum_quda_py_block = ""
        for item in val:
            comment = ""
            idx = enum_quda_h.find(item.name)
            idx_comment = enum_quda_h.find("//", idx)
            idx_endline = enum_quda_h.find("\n", idx)
            if idx_comment != -1 and idx_comment < idx_endline:
                comment = f'\n    """{enum_quda_h[idx_comment + 2 : idx_endline].strip()}"""'
            enum_quda_py_block += f"    {item}{comment}\n"
        if enum_quda_py.find(key) != -1:
            idx_key = enum_quda_py.find(key)
            enum_quda_py = enum_quda_py[:idx_key] + enum_quda_py[idx_key:].replace("    pass\n", enum_quda_py_block, 1)
        else:
            if enum_quda_py.find("class", idx_key) != -1:
                idx_key = enum_quda_py.find("class", idx_key)
                enum_quda_py = (
                    enum_quda_py[:idx_key] + f"class {key}(IntEnum):\n{enum_quda_py_block}\n\n" + enum_quda_py[idx_key:]
                )
                idx_key += len("class")
            else:
                enum_quda_py += f"\n\nclass {key}(IntEnum):\n{enum_quda_py_block}"

    for key, val in quda_params_meta.items():
        quda_pxd_block = ""
        quda_pyx_block = ""
        quda_pyi_block = ""
        for item in val:
            quda_pxd_block += f"        {item}\n"
            if len(item.array) > 0 and item.type.startswith("Quda") and item.type.endswith("Param"):
                quda_pyx_block += multigrid_param.replace("%name%", item.name).replace("%type%", item.type)
                quda_pyi_block += f"    {item.name}: List[{item.type}, {item.array[0]}]\n"
            elif item.type.startswith("Quda") and item.type.endswith("Param"):
                quda_pyx_block += param.replace("%name%", item.name).replace("%type%", item.type)
                quda_pyi_block += f"    {item.name}: {item.type}\n"
            elif len(item.array) == 1:
                if item.type == "char":
                    quda_pyx_block += cstring.replace("%name%", item.name)
                    quda_pyi_block += f"    {item.name}: bytes[{item.array[0]}]\n"
                else:
                    quda_pyx_block += normal.replace("%name%", item.name)
                    quda_pyi_block += f"    {item.name}: List[{item.type}, {item.array[0]}]\n"
            elif len(item.array) == 2:
                quda_pyx_block += multigrid.replace("%name%", item.name)
                if item.type == "char":
                    quda_pyi_block += f"    {item.name}: List[bytes[{item.array[1]}], {item.array[0]}]\n"
                else:
                    quda_pyi_block += f"    {item.name}: List[List[{item.type}, {item.array[1]}], {item.array[0]}]\n"
            elif item.ptr == "*" and item.type == "void":
                quda_pyx_block += void_ptr.replace("%name%", item.name)
                quda_pyi_block += f"    {item.name}: Pointer\n"
            elif item.ptr == "*":
                quda_pyx_block += ptr.replace("%name%", item.name).replace("%type%", item.type)
                quda_pyi_block += f"    {item.name}: Pointer\n"
            elif item.ptr == "**":
                quda_pyx_block += ptrptr.replace("%name%", item.name).replace("%type%", item.type)
                quda_pyi_block += f"    {item.name}: Pointers\n"
            else:
                quda_pyx_block += normal.replace("%name%", item.name)
                quda_pyi_block += f"    {item.name}: {item.type}\n"
        idx = quda_pxd.find(key)
        quda_pxd = quda_pxd[:idx] + quda_pxd[idx:].replace(
            "        pass\n", quda_pxd_block.replace("double _Complex", "double_complex"), 1
        )
        quda_pyx = quda_pyx.replace(
            f"\n##%%!! {key}\n",
            quda_pyx_block.replace("double _Complex", "double_complex"),
        )
        # quda_pyi = quda_pyi.replace(
        #     f"##%%!! {key}\n",
        #     pyi.replace("double _Complex", "double_complex").replace("unsigned int", "int"),
        # )

    with open(os.path.join(pyquda_root, "pyquda", "src", "quda_constants.pxd"), "w") as f:
        f.write(quda_constants_pxd)
    with open(os.path.join(pyquda_root, "pyquda", "src", "enum_quda.pxd"), "w") as f:
        f.write(enum_quda_pxd)
    with open(os.path.join(pyquda_root, "pyquda", "src", "quda.pxd"), "w") as f:
        f.write(quda_pxd)
    with open(os.path.join(pyquda_root, "pyquda", "src", "quda.pyx"), "w") as f:
        f.write(quda_pyx)
    with open(os.path.join(pyquda_root, "pyquda", "enum_quda.py"), "w") as f:
        f.write(enum_quda_py)
    # with open(os.path.join(pyquda_root, "pyquda", "pyquda.pyi"), "w") as f:
    #     f.write(quda_pyi)

    if os.path.exists(os.path.join(pyquda_root, "yacctab.py")):
        os.remove(os.path.join(pyquda_root, "yacctab.py"))
    if os.path.exists(os.path.join(pyquda_root, "lextab.py")):
        os.remove(os.path.join(pyquda_root, "lextab.py"))



from typing import List, Optional, Union

from ..field import LatticeInfo, LatticeGauge, LatticeClover
from ..enum_quda import QudaDslashType

from . import general
from .abstract import Multigrid, FermionDirac


class CloverWilsonDirac(FermionDirac):
    def __init__(
        self,
        latt_info: LatticeInfo,
        mass: float,
        tol: float,
        maxiter: int,
        clover_csw: float = 0.0,
        clover_xi: float = 1.0,
        multigrid: Union[List[List[int]], Multigrid, None] = None,
    ) -> None:
        kappa = 1 / (2 * (mass + 1 + (latt_info.Nd - 1) / latt_info.anisotropy))
        super().__init__(latt_info)
        self.clover: Optional[LatticeClover] = None
        self.clover_inv: Optional[LatticeClover] = None
        self.newQudaGaugeParam()
        self.newQudaMultigridParam(multigrid, mass, kappa)
        self.newQudaInvertParam(mass, kappa, tol, maxiter, clover_csw, clover_xi)
        self.setPrecision()
        self.setReconstruct()

    def newQudaGaugeParam(self):
        gauge_param = general.newQudaGaugeParam(self.latt_info)
        self.gauge_param = gauge_param

    def newQudaMultigridParam(self, multigrid: Union[List[List[int]], Multigrid, None], mass: float, kappa: float):
        if isinstance(multigrid, Multigrid):
            self.multigrid = multigrid
        elif multigrid is not None:
            geo_block_size = multigrid
            mg_param, mg_inv_param = general.newQudaMultigridParam(
                QudaDslashType.QUDA_CLOVER_WILSON_DSLASH, mass, kappa, geo_block_size
            )
            self.multigrid = Multigrid(mg_param, mg_inv_param)
            self.multigrid.setParam()
        else:
            self.multigrid = Multigrid(None, None)

    def newQudaInvertParam(
        self, mass: float, kappa: float, tol: float, maxiter: int, clover_csw: float, clover_xi: float
    ):
        invert_param = general.newQudaInvertParam(
            QudaDslashType.QUDA_CLOVER_WILSON_DSLASH,
            mass,
            kappa,
            tol,
            maxiter,
            kappa * clover_csw,
            clover_xi,
            self.multigrid.param,
        )
        self.invert_param = invert_param

    def saveClover(self, gauge: LatticeGauge):
        if self.clover is None or self.clover_inv is None:
            self.clover = LatticeClover(gauge.latt_info)
            self.clover_inv = LatticeClover(gauge.latt_info)
        general.saveClover(self.clover, self.clover_inv, gauge, self.gauge_param, self.invert_param)

    def restoreClover(self):
        assert self.clover is not None and self.clover_inv is not None
        general.loadClover(self.clover, self.clover_inv, None, self.gauge_param, self.invert_param)
        general.loadMultigrid(self.multigrid, self.invert_param, True)

    def loadGauge(self, gauge: LatticeGauge, thin_update_only: bool = False):
        general.loadClover(self.clover, self.clover_inv, gauge, self.gauge_param, self.invert_param)
        general.loadGauge(gauge, self.gauge_param)
        general.loadMultigrid(self.multigrid, self.invert_param, thin_update_only)

    def freeGauge(self):
        general.freeClover()
        general.freeGauge()
        general.freeMultigrid(self.multigrid, self.invert_param)



from abc import abstractmethod
from math import prod
from os import path
from time import perf_counter
from typing import Any, Dict, Generic, List, Literal, Optional, Sequence, Tuple, Type, TypeVar, Union, overload

Self = TypeVar("Self", bound="BaseField")
SelfMulti = TypeVar("SelfMulti", bound="MultiField")
Field = TypeVar("Field", bound="BaseField")

import numpy
from numpy.lib.format import read_magic, read_array_header_1_0, write_array_header_1_0
from numpy.typing import NDArray

from . import (
    getRankFromCoord,
    getLogger,
    getSublatticeSize,
    getMPIComm,
    getMPISize,
    getMPIRank,
    getGridSize,
    getGridCoord,
    getArrayBackend,
    openReadHeader,
    openWriteHeader,
    readMPIFile,
    writeMPIFile,
)
from .array import (
    BackendType,
    arrayIsInstance,
    arrayDType,
    arrayHost,
    arrayHostCopy,
    arrayDevice,
    arrayCopy,
    arrayIsContiguous,
    arrayAsContiguous,
    arrayLinalgNorm,
    arrayIdentity,
    arrayZeros,
    arrayEmpty,
)
from .pointer import Pointer, ndarrayPointer

_latt_even_odd: Dict[Tuple[int, ...], Tuple[NDArray[numpy.intp], NDArray[numpy.intp]]] = {}
_latt_coord: Dict[Tuple[int, ...], NDArray[numpy.int32]] = {}
_latt_sort_key: Dict[Tuple[int, ...], NDArray[numpy.int64]] = {}
_latt_lexico_coord: Dict[Tuple[int, ...], NDArray[numpy.int32]] = {}
_latt_lexico_sort_key: Dict[Tuple[int, ...], NDArray[numpy.int64]] = {}


class BaseInfo:
    def __init__(self, latt_size: Sequence[int], force_even: bool, Ns: int = 4, Nc: int = 3) -> None:
        self.Nd = len(latt_size)
        self.Ns = Ns
        self.Nc = Nc
        self._setLattice(latt_size, force_even)

    def _setLattice(self, latt_size: Sequence[int], force_even: bool):
        self.mpi_comm = getMPIComm()
        self.mpi_size = getMPISize()
        self.mpi_rank = getMPIRank()
        self.grid_size = getGridSize()
        self.grid_coord = getGridCoord()

        sublatt_size = getSublatticeSize(latt_size, force_even)
        if self.Nd == 4:
            self.Gx, self.Gy, self.Gz, self.Gt = self.grid_size
            self.gx, self.gy, self.gz, self.gt = self.grid_coord
            self.GLx, self.GLy, self.GLz, self.GLt = latt_size
            self.Lx, self.Ly, self.Lz, self.Lt = sublatt_size
        else:
            self.Gt = self.grid_size[-1]
            self.gt = self.grid_coord[-1]
            self.GLt = latt_size[-1]
            self.Lt = sublatt_size[-1]

        self.global_size = [GL for GL in latt_size]
        self.global_volume = prod(latt_size)
        self.size = sublatt_size
        self.volume = prod(sublatt_size)
        self.ga_pad = self.volume // min(sublatt_size) // 2

    @abstractmethod
    def lexico(self, data: NDArray, multi: bool, backend: BackendType = "numpy") -> NDArray:
        getLogger().critical(
            f"{self.__class__.__name__}.lexico(data, multi, backend) not implemented", NotImplementedError
        )

    @abstractmethod
    def coordinate(self, mu: Optional[int] = None) -> NDArray:
        getLogger().critical(f"{self.__class__.__name__}.coordinate(mu) not implemented", NotImplementedError)

    @abstractmethod
    def sortSelect(self, select: NDArray[numpy.intp]) -> NDArray:
        getLogger().critical(f"{self.__class__.__name__}.lexico_index() not implemented", NotImplementedError)


class LatticeInfo(BaseInfo):
    def __init__(
        self, latt_size: List[int], t_boundary: Literal[1, -1] = 1, anisotropy: float = 1.0, Ns: int = 4, Nc: int = 3
    ) -> None:
        super().__init__(latt_size, True, Ns, Nc)
        self.t_boundary = t_boundary
        self.anisotropy = anisotropy

    def _setEvenOdd(self):
        global _latt_even_odd
        key = tuple(self.global_size)
        if key not in _latt_even_odd:
            eo = numpy.sum(numpy.indices(self.size[::-1], "<i4"), axis=0).reshape(-1) % 2
            _latt_even_odd[key] = (numpy.where(eo == 0)[0], numpy.where(eo == 1)[0])
        self._even, self._odd = _latt_even_odd[key]

    def lexico(self, data: NDArray, multi: bool, backend: BackendType = "numpy") -> NDArray:
        self._setEvenOdd()
        shape = data.shape
        if multi:
            L5 = shape[0]
            assert shape[1] == 2
            sublatt_size = list(shape[2 : self.Nd + 2][::-1])
            field_shape = list(shape[self.Nd + 2 :])
        else:
            L5 = 1
            assert shape[0] == 2
            sublatt_size = list(shape[1 : self.Nd + 1][::-1])
            field_shape = list(shape[self.Nd + 1 :])
        sublatt_size[0] *= 2
        assert sublatt_size == self.size
        data_evenodd = data.reshape(L5, 2, self.volume // 2, prod(field_shape))
        data_lexico = arrayEmpty((L5, self.volume, prod(field_shape)), data.dtype, backend)
        data_lexico[:, self._even] = data_evenodd[:, 0]
        data_lexico[:, self._odd] = data_evenodd[:, 1]
        if multi:
            return data_lexico.reshape(L5, *sublatt_size[::-1], *field_shape)
        else:
            return data_lexico.reshape(*sublatt_size[::-1], *field_shape)

    def evenodd(self, data: NDArray, multi: bool, backend: BackendType = "numpy") -> NDArray:
        self._setEvenOdd()
        shape = data.shape
        if multi:
            L5 = shape[0]
            sublatt_size = list(shape[1 : self.Nd + 1][::-1])
            field_shape = list(shape[self.Nd + 1 :])
        else:
            L5 = 1
            sublatt_size = list(shape[0 : self.Nd + 0][::-1])
            field_shape = list(shape[self.Nd + 0 :])
        assert sublatt_size == self.size
        sublatt_size[0] //= 2
        data_lexico = data.reshape(L5, self.volume, prod(field_shape))
        data_evenodd = arrayEmpty((L5, 2, self.volume // 2, prod(field_shape)), data.dtype, backend)
        data_evenodd[:, 0] = data_lexico[:, self._even]
        data_evenodd[:, 1] = data_lexico[:, self._odd]
        if multi:
            return data_evenodd.reshape(L5, 2, *sublatt_size[::-1], *field_shape)
        else:
            return data_evenodd.reshape(2, *sublatt_size[::-1], *field_shape)

    def coordinate(self, mu: Optional[int] = None):
        global _latt_coord
        key = tuple(self.global_size)
        if key not in _latt_coord:
            _latt_coord[key] = self.evenodd(numpy.indices(self.size[::-1], "<i4")[::-1], True)
            for i in range(self.Nd):
                _latt_coord[key][i] += self.grid_coord[i] * self.size[i]
        if mu is None:
            return _latt_coord[key]
        else:
            assert 0 <= mu < self.Nd
            return _latt_coord[key][mu]

    def sortSelect(self, select: NDArray[numpy.intp]) -> NDArray:
        global _latt_sort_key
        key = tuple(self.global_size)
        if key not in _latt_sort_key:
            _latt_sort_key[key] = numpy.array(self.coordinate(0), "<i8")
            for mu in range(1, self.Nd):
                _latt_sort_key[key] *= self.global_size[mu]
                _latt_sort_key[key] += self.coordinate(mu)
        return select[numpy.argsort(_latt_sort_key[key].reshape(-1)[select])]


class LexicoInfo(BaseInfo):
    def __init__(self, latt_size: Sequence[int], Ns: int = 4, Nc: int = 3) -> None:
        super().__init__(latt_size, False, Ns, Nc)

    def lexico(self, data: NDArray, multi: bool, backend: BackendType = "numpy") -> NDArray:
        return arrayCopy(data, backend)

    def coordinate(self, mu: Optional[int] = None):
        global _latt_lexico_coord
        key = tuple(self.global_size)
        if key not in _latt_lexico_coord:
            _latt_lexico_coord[key] = numpy.indices(self.size[::-1], "<i4")[::-1]
            for i in range(self.Nd):
                _latt_lexico_coord[key][i] += self.grid_coord[i] * self.size[i]
        if mu is None:
            return _latt_lexico_coord[key]
        else:
            assert 0 <= mu < self.Nd
            return _latt_lexico_coord[key][mu]

    def sortSelect(self, select: NDArray[numpy.intp]) -> NDArray:
        global _latt_lexico_sort_key
        key = tuple(self.global_size)
        if key not in _latt_lexico_sort_key:
            _latt_lexico_sort_key[key] = numpy.array(self.coordinate(0), "<i8")
            for mu in range(1, self.Nd):
                _latt_lexico_sort_key[key] *= self.global_size[mu]
                _latt_lexico_sort_key[key] += self.coordinate(mu).reshape(-1)
        return select[numpy.argsort(_latt_lexico_sort_key[key][select])]


def read_array_header(filename: str) -> Tuple[Tuple[int, ...], str, int]:
    with openReadHeader(filename) as f:
        if f.fp is not None:
            assert read_magic(f.fp) == (1, 0)
            shape, fortran_order, dtype = read_array_header_1_0(f.fp)
            assert not fortran_order
    return shape, dtype.str, f.offset


def write_array_header(filename: str, shape: Tuple[int, ...], dtype: str) -> int:
    with openWriteHeader(filename) as f:
        if f.fp is not None:
            d = {"shape": shape, "fortran_order": False, "descr": dtype}
            write_array_header_1_0(f.fp, d)
    return f.offset


def _field_spin_color_dtype(
    field: str, shape: Sequence[int], use_fp32: bool
) -> Tuple[Optional[int], Optional[int], str]:
    float_nbytes = 4 if use_fp32 else 8
    if field in ["Int"]:
        () = shape
        return None, None, "<i4"
    elif field in ["Real"]:
        return None, None, f"<f{float_nbytes}"
    elif field in ["Complex"]:
        return None, None, f"<c{2 * float_nbytes}"
    elif field in ["SpinColorVector"]:
        Ns, Nc = shape
        return Ns, Nc, f"<c{2 * float_nbytes}"
    elif field in ["SpinColorMatrix"]:
        Ns, Ns_, Nc, Nc_ = shape
        assert Ns == Ns_ and Nc == Nc_
        return Ns, Nc, f"<c{2 * float_nbytes}"
    elif field in ["SpinVector"]:
        (Ns,) = shape
        return Ns, None, f"<c{2 * float_nbytes}"
    elif field in ["SpinMatrix"]:
        Ns, Ns_ = shape
        assert Ns == Ns_
        return Ns, None, f"<c{2 * float_nbytes}"
    elif field in ["ColorVector"]:
        (Nc,) = shape
        return None, Nc, f"<c{2 * float_nbytes}"
    elif field in ["ColorMatrix"]:
        Nc, Nc_ = shape
        assert Nc == Nc_
        return None, Nc, f"<c{2 * float_nbytes}"
    else:
        getLogger().critical(f"Unknown field type: {field}", ValueError)


def _field_shape_dtype(field: str, Ns: int, Nc: int, use_fp32: bool = False) -> Tuple[List[int], str]:
    float_nbytes = 4 if use_fp32 else 8
    if field in ["Int"]:
        return [], "<i4"
    elif field in ["Real"]:
        return [], f"<f{float_nbytes}"
    elif field in ["Complex"]:
        return [], f"<c{2 * float_nbytes}"
    elif field in ["SpinColorVector"]:
        return [Ns, Nc], f"<c{2 * float_nbytes}"
    elif field in ["SpinColorMatrix"]:
        return [Ns, Ns, Nc, Nc], f"<c{2 * float_nbytes}"
    elif field in ["SpinVector"]:
        return [Ns], f"<c{2 * float_nbytes}"
    elif field in ["SpinMatrix"]:
        return [Ns, Ns], f"<c{2 * float_nbytes}"
    elif field in ["ColorVector"]:
        return [Nc], f"<c{2 * float_nbytes}"
    elif field in ["ColorMatrix"]:
        return [Nc, Nc], f"<c{2 * float_nbytes}"
    elif field in ["Mom"]:
        return [Nc**2 + 1], f"<f{float_nbytes}"
    elif field in ["Clover"]:
        return [2, ((Ns // 2) * Nc) ** 2], f"<f{float_nbytes}"
    else:
        getLogger().critical(f"Unknown field type: {field}", ValueError)


class BaseField:
    def __init__(self, latt_info: BaseInfo, *args, **kwargs) -> None:
        self.latt_info = latt_info
        self._data: NDArray = numpy.empty((0, 0), "<c16")
        self.backend: BackendType = getArrayBackend()
        self.L5: int = 0

    @property
    def location(self) -> BackendType:
        return "numpy" if isinstance(self.data, numpy.ndarray) else self.backend

    @abstractmethod
    def _latticeShape(self) -> Tuple[int, ...]:
        getLogger().critical(f"{self.__class__.__name__}._latticeShape() not implemented", NotImplementedError)

    @classmethod
    def _groupName(cls):
        return (
            cls.__name__.replace("Multi", "")
            .replace("Lexico", "")
            .replace("Link", "ColorMatrix")
            .replace("Rotation", "ColorMatrix")
            .replace("Gauge", "ColorMatrix")
            .replace("StaggeredFermion", "ColorVector")
            .replace("StaggeredPropagator", "ColorMatrix")
            .replace("Fermion", "SpinColorVector")
            .replace("Propagator", "SpinColorMatrix")
        )

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if isinstance(value, numpy.ndarray):
            if self.location == "numpy":
                self._data[:] = value
            else:
                self._data[:] = arrayDevice(value, self.backend)
        elif arrayIsInstance(value, self.backend):
            if self.location == "numpy":
                self._data[:] = arrayHost(value, self.backend)
            else:
                self._data[:] = value
        else:
            self._data[:] = value

    @property
    def data_ptr(self) -> NDArray:
        return self.data.reshape(-1)

    @property
    def data_ptrs(self) -> NDArray:
        if self.L5 == 0:
            getLogger().critical(f"{self.__class__.__name__}.data_ptrs not implemented", NotImplementedError)
        else:
            return self.data.reshape(self.L5, -1)

    @property
    def data_void_ptr(self) -> Pointer:
        return ndarrayPointer(self.data.reshape(-1), True)

    @classmethod
    def _field(cls) -> str:
        group_name = cls._groupName()
        return group_name[group_name.index("Lattice") + len("Lattice") :]

    def _setField(self):
        field_shape, field_dtype = _field_shape_dtype(self._field(), self.latt_info.Ns, self.latt_info.Nc)
        self.field_shape = field_shape
        self.field_dtype = field_dtype
        self.lattice_shape = self._latticeShape()
        self.shape: Tuple[int, ...] = (
            (*self.lattice_shape, *self.field_shape)
            if self.L5 == 0
            else (self.L5, *self.lattice_shape, *self.field_shape)
        )
        self.dtype = arrayDType(field_dtype, self.backend)

    def _initData(self, value: Union[NDArray, BackendType, None]):
        self._setField()
        if value is None:
            self._data = arrayZeros(self.shape, self.dtype, self.backend)
        elif isinstance(value, str):
            self._data = arrayZeros(self.shape, self.dtype, value)
        else:
            location = "numpy" if isinstance(value, numpy.ndarray) else self.backend
            if self.L5 == 0:
                self._data = arrayAsContiguous(value.reshape(self.shape), location)
            else:
                for index in range(self.L5):
                    if not arrayIsContiguous(value[index], location):
                        self._data = arrayAsContiguous(value.reshape(self.shape), location)
                        break
                else:
                    self._data = value.reshape(self.shape)

    def lexico(self, force_numpy: bool = True):
        if not isinstance(self, _FULL_FILED_TUPLE):
            getLogger().critical(f"{self.__class__.__name__}.lexico(force_numpy) not implemented", NotImplementedError)
        if force_numpy:
            return self.latt_info.lexico(self.getHost(), self.L5 > 0)
        else:
            return self.latt_info.lexico(self.data, self.L5 > 0, self.location)

    def _select_lattice(self, coord: Tuple[Union[int, list, tuple, slice], ...]):
        select = numpy.full(self.lattice_shape, True)
        shape = []
        subshape = []
        for i, index in enumerate(coord):
            X = self.latt_info.coordinate(i)
            GL = self.latt_info.global_size[i]
            g = self.latt_info.grid_coord[i]
            L = self.latt_info.size[i]
            subshape.append(0)
            if isinstance(index, int):
                shape.append(1)
                if g * L <= index % GL < (g + 1) * L:
                    subshape[-1] += 1
                select &= X == (index % GL)
            elif isinstance(index, (list, tuple)):
                shape.append(len(index))
                for i in index:
                    if g * L <= i % GL < (g + 1) * L:
                        subshape[-1] += 1
                subselect = X == (index[0] % GL)
                for i in range(1, len(index)):
                    subselect |= X == (index[i] % GL)
                select &= subselect
            elif isinstance(index, slice):
                start, stop, step = index.indices(GL)
                if start >= stop or step <= 0:
                    getLogger().critical(
                        f"{self.__class__.__name__}[...] only accepts a:b:c with a<b and c>0", ValueError
                    )
                shape.append((stop - start + step - 1) // step)
                for i in range(start, stop, step):
                    if g * L <= i < (g + 1) * L:
                        subshape[-1] += 1
                select &= (X >= start) & (X < stop)
                if step > 1:
                    select &= (X - start) % step == 0
            else:
                getLogger().critical(
                    f"{self.__class__.__name__}[...] only accepts "
                    f"int, list, tuple or slice as index, got {type(index)}",
                    ValueError,
                )
        select = numpy.where(select.reshape(-1))[0]
        return self.latt_info.sortSelect(select), shape, subshape

    def _select_field(self, field_coord, field_shape):
        if len(field_coord) == 0:
            return field_shape
        shape = []
        for index, L in zip(field_coord, field_shape):
            if isinstance(index, int):
                pass
            elif isinstance(index, (list, tuple)):
                shape.append(len(index))
            elif isinstance(index, slice):
                start, stop, step = index.indices(L)
                if start >= stop or step <= 0:
                    getLogger().critical(
                        f"{self.__class__.__name__}[...] only accepts a:b:c with a<b and c>0", ValueError
                    )
                shape.append((stop - start + step - 1) // step)
            else:
                getLogger().critical(
                    f"{self.__class__.__name__}[...] only accepts "
                    f"int, list, tuple or slice as index, got {type(index)}",
                    ValueError,
                )
        return shape

    def __setitem__(self, key: Tuple[Union[int, list, tuple, slice], ...], value: Union[int, float, complex, NDArray]):
        Nd = self.latt_info.Nd
        if not isinstance(key, tuple):
            key = (key,)
        if len(key) == Nd or len(key) == Nd + len(self.field_shape):
            select, shape, subshape = self._select_lattice(key[:Nd])
            field_shape = self._select_field(key[Nd:], self.field_shape)
            key_ = (select, *key[Nd:])
            if isinstance(value, (int, float, complex)):
                self.data.reshape(-1, *self.field_shape)[key_] = value
            elif isinstance(value, numpy.ndarray):
                if tuple(field_shape) == numpy.broadcast_shapes(field_shape, value.shape):
                    self.data.reshape(-1, *self.field_shape)[key_] = arrayDevice(value, self.location)
                else:
                    self.data.reshape(-1, *self.field_shape)[key_] = arrayDevice(value, self.location).reshape(
                        *subshape, *field_shape
                    )
            else:
                getLogger().critical(
                    f"{self.__class__.__name__}[...] only accepts "
                    f"int, float, complex or ndarray as value, got {type(value)}",
                    ValueError,
                )
        else:
            getLogger().critical(
                f"{self.__class__.__name__}[...] only accepts "
                f"{Nd} or {Nd + len(self.field_shape)} indices, got {len(key)}",
                ValueError,
            )

    def __getitem__(self, key: Tuple[Union[int, list, tuple, slice], ...]) -> NDArray:
        Nd = self.latt_info.Nd
        if not isinstance(key, tuple):
            key = (key,)
        if len(key) == Nd or len(key) == Nd + len(self.field_shape):
            select, shape, subshape = self._select_lattice(key[:Nd])
            field_shape = self._select_field(key[Nd:], self.field_shape)
            key_ = (select, *key[Nd:])
            return arrayHostCopy(self.data.reshape(-1, *self.field_shape)[key_], self.location).reshape(
                *subshape, *field_shape
            )
        else:
            getLogger().critical(
                f"{self.__class__.__name__}[...] only accepts "
                f"{Nd} or {Nd + len(self.field_shape)} indices, got {len(key)}",
                ValueError,
            )

    @classmethod
    def load(cls: Type[Self], filename: str) -> Self:
        if not issubclass(cls, _FULL_FILED_TUPLE):
            getLogger().critical(f"{cls.__name__}.load(filename) not implemented", NotImplementedError)
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        grid_size = getGridSize()
        Nd = len(grid_size)
        shape, dtype, offset = read_array_header(filename)
        latt_size = list(shape[:Nd][::-1])
        field_shape = list(shape[Nd:])
        sublatt_size = getSublatticeSize(latt_size, False)
        value = readMPIFile(filename, dtype, offset, [*sublatt_size[::-1], *field_shape], list(range(Nd - 1, -1, -1)))
        gbytes += value.nbytes / 1024**3
        Ns, Nc, dtype = _field_spin_color_dtype(cls._field(), field_shape, False)
        value = value.astype(dtype)
        Ns = 4 if Ns is None else Ns
        Nc = 3 if Nc is None else Nc
        if issubclass(cls, LexicoField):
            latt_info = LexicoInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, value)
        elif issubclass(cls, FullField):
            latt_info = LatticeInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, latt_info.evenodd(value, False))
        secs = perf_counter() - s
        getLogger().debug(f"Loaded {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")
        return retval

    def save(self, filename: str, *, use_fp32: bool = False):
        if not isinstance(self, _FULL_FILED_TUPLE):
            getLogger().critical(f"{self.__class__.__name__}.load(filename) not implemented", NotImplementedError)
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        if not filename.endswith(".npy"):
            filename += ".npy"
        field = self.lexico()
        _, _, dtype = _field_spin_color_dtype(self._field(), self.field_shape, use_fp32)
        if self.L5 == 0:
            global_shape = (*self.latt_info.global_size[::-1], *self.field_shape)
            offset = write_array_header(filename, global_shape, dtype)
            shape = (*self.latt_info.size[::-1], *self.field_shape)
            axes = list(range(self.latt_info.Nd - 1, -1, -1))
            writeMPIFile(filename, dtype, offset, shape, axes, field.astype(dtype))
            gbytes += field.nbytes / 1024**3
        else:
            global_shape = (self.L5, *self.latt_info.global_size[::-1], *self.field_shape)
            offset = write_array_header(filename, global_shape, dtype)
            shape = (self.L5, *self.latt_info.size[::-1], *self.field_shape)
            axes = list(range(self.latt_info.Nd, 0, -1))
            writeMPIFile(filename, dtype, offset, shape, axes, field.astype(dtype))
            gbytes += field.nbytes / 1024**3
        secs = perf_counter() - s
        getLogger().debug(f"Saved {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")

    @classmethod
    def loadH5(cls: Type[Self], filename: str, label: Union[int, str], *, check: bool = True) -> Self:
        from .hdf5 import File

        if not issubclass(cls, _FULL_FILED_TUPLE):
            getLogger().critical(f"{cls.__name__}.load(filename, label, **kwargs) not implemented", NotImplementedError)
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        with File(filename, "r") as f:
            latt_size, Ns, Nc, value = f.load(cls._groupName(), label, check=check)
        Ns = 4 if Ns is None else Ns
        Nc = 3 if Nc is None else Nc
        if issubclass(cls, LexicoField):
            latt_info = LexicoInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, value)
        elif issubclass(cls, FullField):
            latt_info = LatticeInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, latt_info.evenodd(value, False))
        secs = perf_counter() - s
        getLogger().debug(f"Loaded {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")
        return retval

    def saveH5(
        self,
        filename: str,
        label: Union[int, str, Sequence[int], Sequence[str]],
        *,
        annotation: str = "",
        check: bool = True,
        use_fp32: bool = False,
    ):
        from .hdf5 import File

        if not isinstance(self, _FULL_FILED_TUPLE):
            getLogger().critical(
                f"{self.__class__.__name__}.save(filename, label, **kwargs) not implemented", NotImplementedError
            )
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        if not filename.endswith(".h5") and not filename.endswith(".hdf5"):
            filename += ".h5"
        with File(filename, "w") as f:
            f.save(self._groupName(), label, self.lexico(), annotation=annotation, check=check, use_fp32=use_fp32)
        secs = perf_counter() - s
        getLogger().debug(f"Saved {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")

    def appendH5(
        self,
        filename: str,
        label: Union[int, str, Sequence[int], Sequence[str]],
        *,
        annotation: str = "",
        check: bool = True,
        use_fp32: bool = False,
    ):
        from .hdf5 import File

        if not isinstance(self, _FULL_FILED_TUPLE):
            getLogger().critical(
                f"{self.__class__.__name__}.append(filename, label, **kwargs) not implemented", NotImplementedError
            )
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        with File(filename, "r+") as f:
            f.append(self._groupName(), label, self.lexico(), annotation=annotation, check=check, use_fp32=use_fp32)
        secs = perf_counter() - s
        getLogger().debug(f"Appended {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")

    def updateH5(
        self,
        filename: str,
        label: Union[int, str, Sequence[int], Sequence[str]],
        *,
        annotation: str = "",
        check: bool = True,
        use_fp32: bool = False,
    ):
        from .hdf5 import File

        if not isinstance(self, _FULL_FILED_TUPLE):
            getLogger().critical(
                f"{self.__class__.__name__}.update(filename, label, **kwargs) not implemented", NotImplementedError
            )
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        with File(filename, "r+") as f:
            f.update(self._groupName(), label, self.lexico(), annotation=annotation, check=check, use_fp32=use_fp32)
        secs = perf_counter() - s
        getLogger().debug(f"Updated {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")

    def copy(self):
        if self.L5 == 0:
            return self.__class__(self.latt_info, arrayCopy(self.data, self.location))
        else:
            return self.__class__(self.latt_info, self.L5, arrayCopy(self.data, self.location))

    def toDevice(self):
        self._data = arrayDevice(self.data, self.backend)

    def toHost(self):
        self._data = arrayHost(self.data, self.location)

    def getHost(self):
        return arrayHostCopy(self.data, self.location)

    def norm2(self, all_reduce=True) -> float:
        norm2 = arrayLinalgNorm(self.data, self.location) ** 2
        if all_reduce:
            return getMPIComm().allreduce(norm2)
        else:
            return norm2

    def __add__(self: Self, rhs: Self) -> Self:
        if not self.__class__ == rhs.__class__:
            return NotImplemented
        assert self.location == rhs.location
        if self.L5 == 0:
            return self.__class__(self.latt_info, self.data + rhs.data)
        else:
            return self.__class__(self.latt_info, self.L5, self.data + rhs.data)

    def __sub__(self: Self, rhs: Self) -> Self:
        if not self.__class__ == rhs.__class__:
            return NotImplemented
        assert self.location == rhs.location
        if self.L5 == 0:
            return self.__class__(self.latt_info, self.data - rhs.data)
        else:
            return self.__class__(self.latt_info, self.L5, self.data - rhs.data)

    def __mul__(self, rhs):
        if self.L5 == 0:
            return self.__class__(self.latt_info, self.data * rhs)
        else:
            return self.__class__(self.latt_info, self.L5, self.data * rhs)

    def __rmul__(self, lhs):
        if self.L5 == 0:
            return self.__class__(self.latt_info, lhs * self.data)
        else:
            return self.__class__(self.latt_info, self.L5, lhs * self.data)

    def __truediv__(self, rhs):
        if self.L5 == 0:
            return self.__class__(self.latt_info, self.data / rhs)
        else:
            return self.__class__(self.latt_info, self.L5, self.data / rhs)

    def __neg__(self):
        if self.L5 == 0:
            return self.__class__(self.latt_info, -self.data)
        else:
            return self.__class__(self.latt_info, self.L5, -self.data)

    def __iadd__(self: Self, rhs: Self) -> Self:
        if not self.__class__ == rhs.__class__:
            return NotImplemented
        assert self.location == rhs.location
        self._data += rhs.data
        return self

    def __isub__(self: Self, rhs: Self) -> Self:
        if not self.__class__ == rhs.__class__:
            return NotImplemented
        assert self.location == rhs.location
        self._data -= rhs.data
        return self

    def __imul__(self, rhs):
        self._data *= rhs
        return self

    def __itruediv__(self, rhs):
        self._data /= rhs
        return self


class LexicoField(BaseField):
    latt_info: LexicoInfo

    def __init__(self, latt_info: LexicoInfo, value: Any = None, init_data: bool = True) -> None:
        super().__init__(latt_info)
        if init_data:
            self._initData(value)

    def _latticeShape(self):
        return tuple(self.latt_info.size[::-1])

    def shift(self, n: int, mu: int):
        assert 0 <= mu < 2 * self.latt_info.Nd
        Nd = self.latt_info.Nd
        direction = 1 if mu < Nd else -1
        direction *= 1 if n >= 0 else -1
        mu = mu % Nd
        n = abs(n)
        location = self.location
        left_slice = [slice(None, None) for _ in range(Nd)]
        right_slice = [slice(None, None) for _ in range(Nd)]
        left = arrayCopy(self.data, location)
        right = arrayCopy(self.data, location) if n > 1 else self.data
        rank = getMPIRank()
        coords = getGridCoord()
        g, G = coords[mu], getGridSize()[mu]
        coords[mu] = (g - direction) % G
        dest = getRankFromCoord(coords)
        coords[mu] = (g + direction) % G
        source = getRankFromCoord(coords)

        if n == 0:
            left, right = right, left
        while n > 0:
            left_slice[mu] = slice(-1, None) if direction == 1 else slice(None, 1)
            right_slice[mu] = slice(None, 1) if direction == 1 else slice(-1, None)
            sendbuf = right[tuple(right_slice[::-1])]
            if rank == source and rank == dest:
                pass
            else:
                sendbuf_host = arrayHostCopy(sendbuf, location)
                request = getMPIComm().Isend(sendbuf_host, dest)

            left_slice[mu] = slice(None, -1) if direction == 1 else slice(1, None)
            right_slice[mu] = slice(1, None) if direction == 1 else slice(None, -1)
            left[tuple(left_slice[::-1])] = right[tuple(right_slice[::-1])]

            left_slice[mu] = slice(-1, None) if direction == 1 else slice(None, 1)
            right_slice[mu] = slice(None, 1) if direction == 1 else slice(-1, None)
            if rank == source and rank == dest:
                recvbuf = sendbuf
            else:
                recvbuf_host = numpy.empty_like(sendbuf_host)
                getMPIComm().Recv(recvbuf_host, source)
                request.Wait()
                recvbuf = arrayDevice(recvbuf_host, location)
            left[tuple(left_slice[::-1])] = recvbuf

            n -= 1
            left, right = right, left

        assert isinstance(self.latt_info, LexicoInfo)
        return self.__class__(self.latt_info, right)


class ParityField(BaseField):
    latt_info: LatticeInfo

    def __init__(self, latt_info: LatticeInfo, value: Any = None, init_data: bool = True) -> None:
        super().__init__(latt_info)
        if init_data:
            self._initData(value)

    def _latticeShape(self):
        return tuple(self.latt_info.size[1:][::-1] + [self.latt_info.size[0] // 2])


class FullField(BaseField, Generic[Field]):
    latt_info: LatticeInfo

    def __init__(self, latt_info: LatticeInfo, value: Any = None, init_data: bool = True) -> None:
        s = super(FullField, self)
        if hasattr(s, "__field_class__"):
            getattr(s, "__field_class__").__base__.__init__(self, latt_info, value, False)
            # s.__field_class__.__base__.__init__(self, latt_info, value, False)
        else:
            s.__init__(latt_info, value, False)
        if init_data:
            self._initData(value)

    def _latticeShape(self):
        return tuple([2] + self.latt_info.size[1:][::-1] + [self.latt_info.size[0] // 2])

    @property
    def even(self) -> Field:
        if self.L5 == 0:
            return getattr(super(FullField, self), "__field_class__")(self.latt_info, self.data[0])
            # return super(FullField, self).__field_class__(self.latt_info, self.data[0])
        else:
            getLogger().critical(f"{self.__class__.__name__}.even not implemented", NotImplementedError)

    @even.setter
    def even(self, value: Field):
        if self.L5 == 0:
            self.data[0] = value.data
        else:
            getLogger().critical(f"{self.__class__.__name__}.even not implemented", NotImplementedError)

    @property
    def even_ptr(self) -> NDArray:
        if self.L5 == 0:
            return self.data.reshape(2, -1)[0]
        else:
            getLogger().critical(f"{self.__class__.__name__}.even_ptr not implemented", NotImplementedError)

    @property
    def even_ptrs(self) -> NDArray:
        if self.L5 == 0:
            getLogger().critical(f"{self.__class__.__name__}.even_ptrs not implemented", NotImplementedError)
        else:
            return self.data.reshape(self.L5, 2, -1)[:, 0]

    @property
    def odd(self) -> Field:
        if self.L5 == 0:
            return getattr(super(FullField, self), "__field_class__")(self.latt_info, self.data[1])
            # return super(FullField, self).__field_class__(self.latt_info, self.data[1])
        else:
            getLogger().critical(f"{self.__class__.__name__}.odd not implemented", NotImplementedError)

    @odd.setter
    def odd(self, value: Field):
        if self.L5 == 0:
            self.data[1] = value.data
        else:
            getLogger().critical(f"{self.__class__.__name__}.odd not implemented", NotImplementedError)

    @property
    def odd_ptr(self) -> NDArray:
        if self.L5 == 0:
            return self.data.reshape(2, -1)[1]
        else:
            getLogger().critical(f"{self.__class__.__name__}.odd_ptr not implemented", NotImplementedError)

    @property
    def odd_ptrs(self) -> NDArray:
        if self.L5 == 0:
            getLogger().critical(f"{self.__class__.__name__}.odd_ptrs not implemented", NotImplementedError)
        else:
            return self.data.reshape(self.L5, 2, -1)[:, 1]

    def shift(self, n: int, mu: int):
        assert 0 <= mu < 2 * self.latt_info.Nd
        Nd = self.latt_info.Nd
        direction = 1 if mu < Nd else -1
        direction *= 1 if n >= 0 else -1
        mu = mu % Nd
        n = abs(n)
        location = self.location
        left_slice = [slice(None, None) for _ in range(Nd)]
        right_slice = [slice(None, None) for _ in range(Nd)]
        left = arrayCopy(self.data, location)
        right = arrayCopy(self.data, location) if n > 1 else self.data
        rank = getMPIRank()
        coords = getGridCoord()
        g, G = coords[mu], getGridSize()[mu]
        coords[mu] = (g - direction) % G
        dest = getRankFromCoord(coords)
        coords[mu] = (g + direction) % G
        source = getRankFromCoord(coords)

        if n == 0:
            left, right = right, left
        while n > 0:
            if mu == 0 and n == 1:
                Lx = self.latt_info.size[0]
                Sx = prod(self.latt_info.size[1:])
                left_flat = left.reshape(2, Sx, Lx // 2, -1)
                right_flat = right.reshape(2, Sx, Lx // 2, -1)
                eo = numpy.sum(numpy.indices((2, *self.latt_info.size[1:][::-1])), axis=0).reshape(-1) % 2
                even = numpy.where(eo == 0)[0].reshape(2, -1) % Sx
                odd = numpy.where(eo == 1)[0].reshape(2, -1) % Sx
                if direction == 1:
                    sendbuf = arrayDevice([right_flat[0, even[0], 0], right_flat[1, even[1], 0]], location)
                    if rank == source and rank == dest:
                        pass
                    else:
                        sendbuf_host = arrayHostCopy(sendbuf, location)
                        request = getMPIComm().Isend(sendbuf_host, dest)

                    left_flat[1, even[1]] = right_flat[0, odd[0]]
                    left_flat[0, even[0]] = right_flat[1, odd[1]]
                    left_flat[1, odd[1], :-1] = right_flat[0, even[0], 1:]
                    left_flat[0, odd[0], :-1] = right_flat[1, even[1], 1:]

                    if rank == source and rank == dest:
                        recvbuf = sendbuf
                    else:
                        recvbuf_host = numpy.empty_like(sendbuf_host)
                        getMPIComm().Recv(recvbuf_host, source)
                        request.Wait()
                        recvbuf = arrayDevice(recvbuf_host, location)
                    left_flat[1, odd[1], -1] = recvbuf[0]
                    left_flat[0, odd[0], -1] = recvbuf[1]
                else:
                    sendbuf = arrayDevice([right_flat[0, odd[0], -1], right_flat[1, odd[1], -1]], location)
                    if rank == source and rank == dest:
                        pass
                    else:
                        sendbuf_host = arrayHostCopy(sendbuf, location)
                        request = getMPIComm().Isend(sendbuf_host, dest)

                    left_flat[1, odd[1]] = right_flat[0, even[0]]
                    left_flat[0, odd[0]] = right_flat[1, even[1]]
                    left_flat[1, even[1], 1:] = right_flat[0, odd[0], :-1]
                    left_flat[0, even[0], 1:] = right_flat[1, odd[1], :-1]

                    if rank == source and rank == dest:
                        recvbuf = sendbuf
                    else:
                        recvbuf_host = numpy.empty_like(sendbuf_host)
                        getMPIComm().Recv(recvbuf_host, source)
                        request.Wait()
                        recvbuf = arrayDevice(recvbuf_host, location)
                    left_flat[1, even[1], 0] = recvbuf[0]
                    left_flat[0, even[0], 0] = recvbuf[1]

                n -= 1
            else:
                left_slice[mu] = slice(-1, None) if direction == 1 else slice(None, 1)
                right_slice[mu] = slice(None, 1) if direction == 1 else slice(-1, None)
                sendbuf = right[(slice(None, None),) + tuple(right_slice[::-1])]
                if rank == source and rank == dest:
                    pass
                else:
                    sendbuf_host = arrayHostCopy(sendbuf, location)
                    request = getMPIComm().Isend(sendbuf_host, dest)

                left_slice[mu] = slice(None, -1) if direction == 1 else slice(1, None)
                right_slice[mu] = slice(1, None) if direction == 1 else slice(None, -1)
                if mu == 0:
                    left[(0,) + tuple(left_slice[::-1])] = right[(0,) + tuple(right_slice[::-1])]
                    left[(1,) + tuple(left_slice[::-1])] = right[(1,) + tuple(right_slice[::-1])]
                else:
                    left[(0,) + tuple(left_slice[::-1])] = right[(1,) + tuple(right_slice[::-1])]
                    left[(1,) + tuple(left_slice[::-1])] = right[(0,) + tuple(right_slice[::-1])]

                left_slice[mu] = slice(-1, None) if direction == 1 else slice(None, 1)
                right_slice[mu] = slice(None, 1) if direction == 1 else slice(-1, None)
                if rank == source and rank == dest:
                    recvbuf = sendbuf
                else:
                    recvbuf_host = numpy.empty_like(sendbuf_host)
                    getMPIComm().Recv(recvbuf_host, source)
                    request.Wait()
                    recvbuf = arrayDevice(recvbuf_host, location)
                if mu == 0:
                    left[(0,) + tuple(left_slice[::-1])] = recvbuf[0]
                    left[(1,) + tuple(left_slice[::-1])] = recvbuf[1]
                else:
                    left[(0,) + tuple(left_slice[::-1])] = recvbuf[1]
                    left[(1,) + tuple(left_slice[::-1])] = recvbuf[0]

                if mu == 0:
                    n -= 2
                else:
                    n -= 1
            left, right = right, left

        return self.__class__(self.latt_info, right)


class MultiField(BaseField, Generic[Field]):
    def __init__(self, latt_info: BaseInfo, L5: int, value: Any = None, init_data: bool = True) -> None:
        s = super(MultiField, self)
        if hasattr(s, "__field_class__"):
            getattr(s, "__field_class__").__base__.__init__(self, latt_info, value, False)
            # s.__field_class__.__base__.__init__(self, latt_info, value, False)
        else:
            s.__init__(latt_info, value, False)
        assert L5 > 0
        self.L5 = L5
        if init_data:
            self._initData(value)

    @overload
    def __getitem__(self, key: int) -> Field: ...
    @overload
    def __getitem__(self: SelfMulti, key: Union[list, tuple, slice]) -> SelfMulti: ...

    def __getitem__(self: SelfMulti, key: Union[int, list, tuple, slice]) -> Union[Field, SelfMulti]:
        if isinstance(key, int):
            return getattr(super(MultiField, self), "__field_class__")(self.latt_info, self.data[key])
            # return super(MultiField, self).__field_class__(self.latt_info, self.data[key])
        elif isinstance(key, list):
            return self.__class__(self.latt_info, len(key), self.data[key])
        elif isinstance(key, tuple):
            return self.__class__(self.latt_info, len(key), self.data[list(key)])
        elif isinstance(key, slice):
            return self.__class__(self.latt_info, len(range(*key.indices(self.L5))), self.data[key])

    def __setitem__(self, key: Union[int, list, tuple, slice], value: Field):
        self.data[key] = value.data

    def shift(self: SelfMulti, n: Sequence[int], mu: Sequence[int]) -> SelfMulti:
        ret = self.copy()
        for i in range(self.L5):
            ret[i] = self[i].shift(n[i], mu[i])
        return ret

    @classmethod
    def load(cls: Type[SelfMulti], filename: str) -> SelfMulti:
        if not issubclass(cls, _FULL_FILED_TUPLE):
            getLogger().critical(f"{cls.__name__}.load(filename) not implemented", NotImplementedError)
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        grid_size = getGridSize()
        Nd = len(grid_size)
        shape, dtype, offset = read_array_header(filename)
        L5 = shape[0]
        latt_size = list(shape[1 : Nd + 1][::-1])
        field_shape = list(shape[Nd + 1 :])
        sublatt_size = getSublatticeSize(latt_size, False)
        value = readMPIFile(filename, dtype, offset, [L5, *sublatt_size[::-1], *field_shape], list(range(Nd, 0, -1)))
        gbytes += value.nbytes / 1024**3
        Ns, Nc, dtype = _field_spin_color_dtype(cls._field(), field_shape, False)
        value = value.astype(dtype)
        Ns = 4 if Ns is None else Ns
        Nc = 3 if Nc is None else Nc
        if issubclass(cls, LexicoField):
            latt_info = LexicoInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, L5, value)
        elif issubclass(cls, FullField):
            latt_info = LatticeInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, L5, latt_info.evenodd(value, True))
        secs = perf_counter() - s
        getLogger().debug(f"Loaded {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")
        return retval

    @classmethod
    def loadH5(
        cls: Type[SelfMulti], filename: str, label: Union[Sequence[int], Sequence[str]], *, check: bool = True
    ) -> SelfMulti:
        from .hdf5 import File

        if not issubclass(cls, _FULL_FILED_TUPLE):
            getLogger().critical(f"{cls.__name__}.load(filename, label, **kwargs) not implemented", NotImplementedError)
        s = perf_counter()
        gbytes = 0
        filename = path.expanduser(path.expandvars(filename))
        with File(filename, "r") as f:
            latt_size, Ns, Nc, value = f.load(cls._groupName(), label, check=check)
        Ns = 4 if Ns is None else Ns
        Nc = 3 if Nc is None else Nc
        if issubclass(cls, LexicoField):
            latt_info = LexicoInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, len(label), value)
        elif issubclass(cls, FullField):
            latt_info = LatticeInfo(latt_size, Ns=Ns, Nc=Nc)
            retval = cls(latt_info, len(label), latt_info.evenodd(value, True))
        secs = perf_counter() - s
        getLogger().debug(f"Loaded {filename} in {secs:.3f} secs, {gbytes / secs:.3f} GB/s")
        return retval


_FULL_FILED_TUPLE = (LexicoField, FullField)


class LatticeInt(FullField, ParityField):
    @property
    def __field_class__(self):
        return LatticeInt


class MultiLatticeInt(MultiField[LatticeInt], LatticeInt):
    pass


class LatticeReal(FullField, ParityField):
    @property
    def __field_class__(self):
        return LatticeReal


class MultiLatticeReal(MultiField[LatticeReal], LatticeReal):
    pass


class LatticeComplex(FullField, ParityField):
    @property
    def __field_class__(self):
        return LatticeComplex


class MultiLatticeComplex(MultiField[LatticeComplex], LatticeComplex):
    pass


class LatticeLink(FullField, ParityField):
    def __init__(self, latt_info: LatticeInfo, value=None) -> None:
        super().__init__(latt_info, value)
        if value is None:
            self.data[:] = arrayIdentity(latt_info.Nc, self.dtype, self.location)

    @property
    def __field_class__(self):
        return LatticeLink


class LatticeRotation(MultiField[LatticeLink], LatticeLink):
    def __init__(self, latt_info: LatticeInfo, value=None) -> None:
        super().__init__(latt_info, 1, value)
        if value is None:
            self.data[:] = arrayIdentity(latt_info.Nc, self.dtype, self.location)


class LatticeGauge(MultiField[LatticeLink], LatticeLink):
    def __init__(self, latt_info: LatticeInfo, L5: int, value=None) -> None:
        super().__init__(latt_info, L5, value)
        if value is None:
            self.data[:] = arrayIdentity(latt_info.Nc, self.dtype, self.location)

    def setAntiPeriodicT(self):
        if self.latt_info.gt == self.latt_info.Gt - 1:
            self.data[-1, :, -1] *= -1

    def setAnisotropy(self, anisotropy: float):
        self.data[:-1] /= anisotropy


class LatticeMom(MultiField, FullField, ParityField):
    pass


class LatticeClover(FullField, ParityField):
    pass


class HalfLatticeFermion(ParityField):
    @property
    def __field_class__(self):
        return HalfLatticeFermion


class MultiHalfLatticeFermion(MultiField[HalfLatticeFermion], HalfLatticeFermion):
    pass


class LatticeFermion(FullField[HalfLatticeFermion], HalfLatticeFermion):
    @property
    def __field_class__(self):
        return LatticeFermion


class MultiLatticeFermion(MultiField[LatticeFermion], LatticeFermion):
    pass


class LatticePropagator(FullField, ParityField):
    def setFermion(self, fermion: LatticeFermion, spin: int, color: int):
        self.data[:, :, :, :, :, :, spin, :, color] = fermion.data

    def getFermion(self, spin: int, color: int):
        return LatticeFermion(self.latt_info, self.data[:, :, :, :, :, :, spin, :, color])


class HalfLatticeStaggeredFermion(ParityField):
    @property
    def __field_class__(self):
        return HalfLatticeStaggeredFermion


class MultiHalfLatticeStaggeredFermion(MultiField[HalfLatticeStaggeredFermion], HalfLatticeStaggeredFermion):
    pass


class LatticeStaggeredFermion(FullField[HalfLatticeStaggeredFermion], HalfLatticeStaggeredFermion):
    @property
    def __field_class__(self):
        return LatticeStaggeredFermion


class MultiLatticeStaggeredFermion(MultiField[LatticeStaggeredFermion], LatticeStaggeredFermion):
    pass


class LatticeStaggeredPropagator(FullField, ParityField):
    def setFermion(self, fermion: LatticeStaggeredFermion, color: int):
        self.data[:, :, :, :, :, :, color] = fermion.data

    def getFermion(self, color: int):
        return LatticeStaggeredFermion(self.latt_info, self.data[:, :, :, :, :, :, color])



from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, List, Dict, Tuple, Optional

from pyquda_comm import getLogger
from pyquda_comm.field import (
    LatticeInfo,
    LatticeGauge,
    LatticeFermion,
    MultiLatticeFermion,
    LatticeStaggeredFermion,
    MultiLatticeStaggeredFermion,
)
from ..quda import (
    QudaGaugeParam,
    QudaInvertParam,
    QudaEigParam,
    QudaGaugeSmearParam,
    QudaGaugeObservableParam,
    invertQuda,
    invertMultiSrcQuda,
    MatQuda,
    MatDagMatQuda,
    dslashQuda,
    dslashMultiSrcQuda,
    cloverQuda,
)
from ..enum_quda import (
    QUDA_MAX_MG_LEVEL,
    QudaParity,
    QudaPrecision,
    QudaReconstructType,
    QudaSolutionType,
    QudaMatPCType,
    QudaSolverNormalization,
    QudaVerbosity,
)

from .general import (
    Precision,
    Reconstruct,
    Multigrid,
    getGlobalPrecision,
    getGlobalReconstruct,
    setPrecisionParam,
    setReconstructParam,
)


class DiracGaugeStack:
    loaded: List[bool]
    stack: List[Tuple["Dirac", LatticeGauge, Dict[str, Any]]]

    def __init__(self):
        self.loaded = []
        self.stack = []

    def load(self, dirac: "Dirac", gauge: LatticeGauge, kwargs: Dict[str, Any]):
        if len(self.stack) > 0 and self.stack[-1][0] is dirac and self.stack[-1][1] is gauge:
            self.loaded.append(False)
        else:
            self.loaded.append(True)
            self.stack.append((dirac, gauge, kwargs))
            dirac.loadGauge(gauge, **kwargs)

    def free(self):
        if self.loaded.pop():
            dirac, gauge, kwargs = self.stack.pop()
            dirac.freeGauge()
            if len(self.loaded) > 0 and self.loaded[-1]:
                dirac, gauge, kwargs = self.stack[-1]
                dirac.loadGauge(gauge, **kwargs)


_DIRAC_GAUGE_STACK = DiracGaugeStack()


class Dirac(ABC):
    latt_info: LatticeInfo
    precision: Precision
    reconstruct: Reconstruct
    gauge_param: QudaGaugeParam
    invert_param: QudaInvertParam
    eig_param: QudaEigParam
    smear_param: QudaGaugeSmearParam
    obs_param: QudaGaugeObservableParam

    def __init__(self, latt_info: LatticeInfo) -> None:
        self.latt_info = latt_info
        self.precision = getGlobalPrecision("none")
        self.reconstruct = getGlobalReconstruct("none")

    @abstractmethod
    def loadGauge(self, gauge: LatticeGauge):
        raise NotImplementedError()

    @abstractmethod
    def freeGauge(self):
        raise NotImplementedError()

    @contextmanager
    def useGauge(self, gauge: LatticeGauge):
        try:
            _DIRAC_GAUGE_STACK.load(self, gauge, {})
            yield self
        finally:
            _DIRAC_GAUGE_STACK.free()

    def setPrecision(
        self,
        *,
        cuda: Optional[QudaPrecision] = None,
        sloppy: Optional[QudaPrecision] = None,
        refinement_sloppy: Optional[QudaPrecision] = None,
        precondition: Optional[QudaPrecision] = None,
        eigensolver: Optional[QudaPrecision] = None,
    ):
        self.precision = Precision(
            cuda if cuda is not None else self.precision.cuda,
            sloppy if sloppy is not None else self.precision.sloppy,
            refinement_sloppy if refinement_sloppy is not None else self.precision.refinement_sloppy,
            precondition if precondition is not None else self.precision.precondition,
            eigensolver if eigensolver is not None else self.precision.eigensolver,
        )
        setPrecisionParam(self.precision, self.gauge_param, self.invert_param, None, None)

    def setReconstruct(
        self,
        *,
        cuda: Optional[QudaReconstructType] = None,
        sloppy: Optional[QudaReconstructType] = None,
        refinement_sloppy: Optional[QudaReconstructType] = None,
        precondition: Optional[QudaReconstructType] = None,
        eigensolver: Optional[QudaReconstructType] = None,
    ):
        self.reconstruct = Reconstruct(
            cuda if cuda is not None else self.reconstruct.cuda,
            sloppy if sloppy is not None else self.reconstruct.sloppy,
            refinement_sloppy if refinement_sloppy is not None else self.reconstruct.refinement_sloppy,
            precondition if precondition is not None else self.reconstruct.precondition,
            eigensolver if eigensolver is not None else self.reconstruct.eigensolver,
        )
        setReconstructParam(self.reconstruct, self.gauge_param)

    def setVerbosity(self, verbosity: QudaVerbosity):
        self.invert_param.verbosity = verbosity
        self.invert_param.verbosity_precondition = verbosity


class FermionDirac(Dirac):
    multigrid: Multigrid

    def __init__(self, latt_info: LatticeInfo) -> None:
        super().__init__(latt_info)
        self.reconstruct = getGlobalReconstruct("wilson")

    @abstractmethod
    def loadGauge(self, gauge: LatticeGauge, thin_update_only: bool = False):
        raise NotImplementedError()

    @abstractmethod
    def freeGauge(self):
        raise NotImplementedError()

    @contextmanager
    def useGauge(self, gauge: LatticeGauge):
        try:
            _DIRAC_GAUGE_STACK.load(self, gauge, {"thin_update_only": True})
            yield self
        finally:
            _DIRAC_GAUGE_STACK.free()

    def setPrecision(
        self,
        multishift: bool = False,
        *,
        cuda: Optional[QudaPrecision] = None,
        sloppy: Optional[QudaPrecision] = None,
        refinement_sloppy: Optional[QudaPrecision] = None,
        precondition: Optional[QudaPrecision] = None,
        eigensolver: Optional[QudaPrecision] = None,
    ):
        if multishift:  # Use SINGLE as sloppy and HALF as refinement_sloppy
            self.precision = getGlobalPrecision("multishift")
        elif self.multigrid.param is not None and self.multigrid.inv_param is not None:
            self.precision = getGlobalPrecision("multigrid")
        else:
            self.precision = getGlobalPrecision("invert")
        super().setPrecision(
            cuda=cuda,
            sloppy=sloppy,
            refinement_sloppy=refinement_sloppy,
            precondition=precondition,
            eigensolver=eigensolver,
        )
        setPrecisionParam(self.precision, None, None, self.multigrid.param, self.multigrid.inv_param)

    def setVerbosity(self, verbosity: QudaVerbosity):
        super().setVerbosity(verbosity)
        if self.multigrid.param is not None and self.multigrid.inv_param is not None:
            self.multigrid.param.verbosity = [verbosity] * QUDA_MAX_MG_LEVEL
            self.multigrid.inv_param.verbosity = verbosity

    def performance(self, tag: str = ""):
        gflops, secs = self.invert_param.gflops, self.invert_param.secs
        if self.invert_param.verbosity >= QudaVerbosity.QUDA_SUMMARIZE:
            if tag == "":
                getLogger().info(f"Time = {secs:.3f} secs, Performance = {gflops / secs:.3f} GFLOPS")
            else:
                getLogger().info(f"Time = {secs:.3f} secs, Performance = {gflops / secs:.3f} GFLOPS ({tag})")

    def invert(self, b: LatticeFermion):
        x = LatticeFermion(b.latt_info)
        invertQuda(x.data_ptr, b.data_ptr, self.invert_param)
        self.performance()
        return x

    def invertRestart(self, b: LatticeFermion, restart: int):
        x = self.invert(b)
        if restart > 0:
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_SOURCE_NORMALIZATION
            x += self.invertRestart(b - self.mat(x), restart - 1)
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_DEFAULT_NORMALIZATION
        return x

    def mat(self, x: LatticeFermion):
        b = LatticeFermion(x.latt_info)
        MatQuda(b.data_ptr, x.data_ptr, self.invert_param)
        return b

    def matDagMat(self, x: LatticeFermion):
        b = LatticeFermion(x.latt_info)
        MatDagMatQuda(b.data_ptr, x.data_ptr, self.invert_param)
        return b

    def dslash(self, x: LatticeFermion, parity: QudaParity = QudaParity.QUDA_INVALID_PARITY):
        b = LatticeFermion(x.latt_info)
        if parity == QudaParity.QUDA_EVEN_PARITY:
            dslashQuda(b.even_ptr, x.odd_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
        elif parity == QudaParity.QUDA_ODD_PARITY:
            dslashQuda(b.odd_ptr, x.even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        elif parity == QudaParity.QUDA_INVALID_PARITY:
            dslashQuda(b.even_ptr, x.odd_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
            dslashQuda(b.odd_ptr, x.even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        return b

    def invertMultiSrc(self, b: MultiLatticeFermion):
        self.invert_param.num_src = b.L5
        x = MultiLatticeFermion(b.latt_info, b.L5)
        invertMultiSrcQuda(x.data_ptrs, b.data_ptrs, self.invert_param)
        self.performance()
        return x

    def invertMultiSrcRestart(self, b: MultiLatticeFermion, restart: int):
        x = self.invertMultiSrc(b)
        if restart > 0:
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_SOURCE_NORMALIZATION
            x += self.invertMultiSrcRestart(b - self.matMultiSrc(x), restart - 1)
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_DEFAULT_NORMALIZATION
        return x

    def matMultiSrc(self, x: MultiLatticeFermion):
        b = MultiLatticeFermion(x.latt_info, x.L5)
        for i in range(x.L5):
            MatQuda(b[i].data_ptr, x[i].data_ptr, self.invert_param)
        return b

    def matDagMatMultiSrc(self, x: MultiLatticeFermion):
        b = MultiLatticeFermion(x.latt_info, x.L5)
        for i in range(x.L5):
            MatDagMatQuda(b[i].data_ptr, x[i].data_ptr, self.invert_param)
        return b

    def dslashMultiSrc(self, x: MultiLatticeFermion, parity: QudaParity = QudaParity.QUDA_INVALID_PARITY):
        self.invert_param.num_src = x.L5
        b = MultiLatticeFermion(x.latt_info, x.L5)
        if parity == QudaParity.QUDA_EVEN_PARITY:
            dslashMultiSrcQuda(b.even_ptrs, x.odd_ptrs, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
        elif parity == QudaParity.QUDA_ODD_PARITY:
            dslashMultiSrcQuda(b.odd_ptrs, x.even_ptrs, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        elif parity == QudaParity.QUDA_INVALID_PARITY:
            dslashMultiSrcQuda(b.even_ptrs, x.odd_ptrs, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
            dslashMultiSrcQuda(b.odd_ptrs, x.even_ptrs, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        return b

    def invertPC(self, b: LatticeFermion):
        self.invert_param.solution_type = QudaSolutionType.QUDA_MATPC_SOLUTION
        self.invert_param.matpc_type = QudaMatPCType.QUDA_MATPC_ODD_ODD

        kappa = self.invert_param.kappa
        x = LatticeFermion(b.latt_info)
        dslashQuda(x.odd_ptr, b.even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        x.even = b.odd + kappa * x.odd
        # * QUDA_ASYMMETRIC_MASS_NORMALIZATION makes the even part 1 / (2 * kappa) instead of 1
        invertQuda(x.odd_ptr, x.even_ptr, self.invert_param)
        self.performance()
        dslashQuda(x.even_ptr, x.odd_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
        x.even = kappa * (2 * b.even + x.even)

        self.invert_param.solution_type = QudaSolutionType.QUDA_MAT_SOLUTION
        self.invert_param.matpc_type = QudaMatPCType.QUDA_MATPC_EVEN_EVEN
        return x

    def invertCloverPC(self, b: LatticeFermion):
        a = LatticeFermion(b.latt_info)
        cloverQuda(a.even_ptr, b.even_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY, 1)
        cloverQuda(a.odd_ptr, b.odd_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY, 1)
        return self.invertPC(a)


class StaggeredFermionDirac(FermionDirac):
    def __init__(self, latt_info: LatticeInfo) -> None:
        super().__init__(latt_info)
        self.reconstruct = getGlobalReconstruct("none")

    def invert(self, b: LatticeStaggeredFermion):
        x = LatticeStaggeredFermion(b.latt_info)
        invertQuda(x.data_ptr, b.data_ptr, self.invert_param)
        self.performance()
        return x

    def invertRestart(self, b: LatticeStaggeredFermion, restart: int):
        x = self.invert(b)
        if restart > 0:
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_SOURCE_NORMALIZATION
            x += self.invertRestart(b - self.mat(x), restart - 1)
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_DEFAULT_NORMALIZATION
        return x

    def mat(self, x: LatticeStaggeredFermion):
        b = LatticeStaggeredFermion(x.latt_info)
        MatQuda(b.data_ptr, x.data_ptr, self.invert_param)
        return b

    def matDagMat(self, x: LatticeStaggeredFermion):
        b = LatticeStaggeredFermion(x.latt_info)
        MatDagMatQuda(b.data_ptr, x.data_ptr, self.invert_param)
        return b

    def dslash(self, x: LatticeStaggeredFermion, parity: QudaParity = QudaParity.QUDA_INVALID_PARITY):
        b = LatticeStaggeredFermion(x.latt_info)
        if parity == QudaParity.QUDA_EVEN_PARITY:
            dslashQuda(b.even_ptr, x.odd_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
        elif parity == QudaParity.QUDA_ODD_PARITY:
            dslashQuda(b.odd_ptr, x.even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        elif parity == QudaParity.QUDA_INVALID_PARITY:
            dslashQuda(b.even_ptr, x.odd_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
            dslashQuda(b.odd_ptr, x.even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        return b

    def invertMultiSrc(self, b: MultiLatticeStaggeredFermion):
        self.invert_param.num_src = b.L5
        x = MultiLatticeStaggeredFermion(b.latt_info, b.L5)
        invertMultiSrcQuda(x.data_ptrs, b.data_ptrs, self.invert_param)
        self.performance()
        return x

    def invertMultiSrcRestart(self, b: MultiLatticeStaggeredFermion, restart: int):
        x = self.invertMultiSrc(b)
        if restart > 0:
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_SOURCE_NORMALIZATION
            x += self.invertMultiSrcRestart(b - self.matMultiSrc(x), restart - 1)
            self.invert_param.solver_normalization = QudaSolverNormalization.QUDA_DEFAULT_NORMALIZATION
        return x

    def matMultiSrc(self, x: MultiLatticeStaggeredFermion):
        b = MultiLatticeStaggeredFermion(x.latt_info, x.L5)
        for i in range(x.L5):
            MatQuda(b[i].data_ptr, x[i].data_ptr, self.invert_param)
        return b

    def matDagMatMultiSrc(self, x: MultiLatticeStaggeredFermion):
        b = MultiLatticeStaggeredFermion(x.latt_info, x.L5)
        for i in range(x.L5):
            MatDagMatQuda(b[i].data_ptr, x[i].data_ptr, self.invert_param)
        return b

    def dslashMultiSrc(self, x: MultiLatticeStaggeredFermion, parity: QudaParity = QudaParity.QUDA_INVALID_PARITY):
        self.invert_param.num_src = x.L5
        b = MultiLatticeStaggeredFermion(x.latt_info, x.L5)
        if parity == QudaParity.QUDA_EVEN_PARITY:
            dslashMultiSrcQuda(b.even_ptrs, x.odd_ptrs, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
        elif parity == QudaParity.QUDA_ODD_PARITY:
            dslashMultiSrcQuda(b.odd_ptrs, x.even_ptrs, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        elif parity == QudaParity.QUDA_INVALID_PARITY:
            dslashMultiSrcQuda(b.even_ptrs, x.odd_ptrs, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
            dslashMultiSrcQuda(b.odd_ptrs, x.even_ptrs, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        return b

    def invertPC(self, b: LatticeStaggeredFermion):
        self.invert_param.solution_type = QudaSolutionType.QUDA_MATPC_SOLUTION
        self.invert_param.matpc_type = QudaMatPCType.QUDA_MATPC_ODD_ODD

        mass = self.invert_param.mass
        x = LatticeStaggeredFermion(b.latt_info)
        dslashQuda(x.odd_ptr, b.even_ptr, self.invert_param, QudaParity.QUDA_ODD_PARITY)
        x.even = (2 * mass) * b.odd + x.odd
        invertQuda(x.odd_ptr, x.even_ptr, self.invert_param)
        self.performance()
        dslashQuda(x.even_ptr, x.odd_ptr, self.invert_param, QudaParity.QUDA_EVEN_PARITY)
        x.even = (0.5 / mass) * (b.even + x.even)

        self.invert_param.solution_type = QudaSolutionType.QUDA_MAT_SOLUTION
        self.invert_param.matpc_type = QudaMatPCType.QUDA_MATPC_EVEN_EVEN
        return x



from argparse import ArgumentParser

from . import init


def main():
    parser = ArgumentParser(prog="pyquda", description="PyQUDA initializer", epilog="")
    parser.add_argument("script")
    parser.add_argument(
        "-g",
        "--grid",
        "--geom",
        nargs=4,
        type=int,
        help="GPU grid size used to split the lattice",
        metavar=("Gx", "Gy", "Gz", "Gt"),
    )
    parser.add_argument(
        "-l",
        "--latt",
        nargs=4,
        type=int,
        help="Lattice size used to set the default GPU grid",
        metavar=("Lx", "Ly", "Lz", "Lt"),
    )
    parser.add_argument(
        "-m",
        "--grid-map",
        default="default",
        choices=("default", "reversed", "shared"),
        help="Grid mapping used for PyQUDA (default: default)",
    )
    parser.add_argument(
        "-b",
        "--backend",
        default="cupy",
        choices=("numpy", "cupy", "torch", "dpnp"),
        help="Array backend used for PyQUDA (default: cupy)",
    )
    parser.add_argument(
        "-t",
        "--backend-target",
        default="cuda",
        choices=("cpu", "cuda", "hip", "sycl"),
        help="Array backend target used for PyQUDA (default: cuda)",
    )
    parser.add_argument(
        "--no-init-quda",
        action="store_true",
        help="Don't initialize the QUDA library",
    )
    parser.add_argument(
        "--use-malloc_quda",
        action="store_true",
        help="Use QUDA's allocator for array backends",
    )
    parser.add_argument(
        "-p",
        "--resource-path",
        help="(default: QUDA_RESOURCE_PATH)",
        metavar="QUDA_RESOURCE_PATH",
    )
    args = parser.parse_args()
    init(
        args.grid,
        args.latt,
        args.grid_map,
        args.backend,
        args.backend_target,
        not args.no_init_quda,
        args.use_malloc_quda,
        resource_path=args.resource_path,
    )
    exec(open(args.script).read(), globals(), globals())


if __name__ == "__main__":
    main()



#pragma once

#
# @file  quda.h
# @brief Main header file for the QUDA library
#
# Note to QUDA developers: When adding new members to QudaGaugeParam
# and QudaInvertParam, be sure to update lib/check_params.h as well
# as the Fortran interface in lib/quda_fortran.F90.
#

from libcpp cimport bool
from enum_quda cimport *
from libc.stdio cimport FILE
# from quda_define cimport *
from quda_constants cimport *

cdef extern from "complex.h":
    pass
ctypedef double complex double_complex

cdef extern from "quda.h":

    #
    # Parameters having to do with the gauge field or the
    # interpretation of the gauge field by various Dirac operators
    #
    ctypedef struct QudaGaugeParam:
        pass

    #
    # Parameters relating to the solver and the choice of Dirac operator.
    #
    ctypedef struct QudaInvertParam:
        pass

    # Parameter set for solving eigenvalue problems.
    ctypedef struct QudaEigParam:
        pass

    ctypedef struct QudaMultigridParam:
        pass

    ctypedef struct QudaGaugeObservableParam:
        pass

    ctypedef struct QudaGaugeSmearParam:
        pass

    ctypedef struct QudaBLASParam:
        pass

    #
    # Interface functions, found in interface_quda.cpp
    #

    #
    # Set parameters related to status reporting.
    #
    # In typical usage, this function will be called once (or not at
    # all) just before the call to initQuda(), but it's valid to call
    # it any number of times at any point during execution.  Prior to
    # the first time it's called, the parameters take default values
    # as indicated below.
    #
    # @param verbosity  Default verbosity, ranging from QUDA_SILENT to
    #                   QUDA_DEBUG_VERBOSE.  Within a solver, this
    #                   parameter is overridden by the "verbosity"
    #                   member of QudaInvertParam.  The default value
    #                   is QUDA_SUMMARIZE.
    #
    # @param prefix     String to prepend to all messages from QUDA.  This
    #                   defaults to the empty string (""), but you may
    #                   wish to specify something like "QUDA: " to
    #                   distinguish QUDA's output from that of your
    #                   application.
    #
    # @param outfile    File pointer (such as stdout, stderr, or a handle
    #                   returned by fopen()) where messages should be
    #                   printed.  The default is stdout.
    #
    void setVerbosityQuda(QudaVerbosity verbosity, const char prefix[],
                          FILE *outfile)

    #
    # initCommsGridQuda() takes an optional "rank_from_coords" argument that
    # should be a pointer to a user-defined function with this prototype.
    #
    # @param coords  Node coordinates
    # @param fdata   Any auxiliary data needed by the function
    # @return        MPI rank or QMP node ID cooresponding to the node coordinates
    #
    # @see initCommsGridQuda
    #
    ctypedef int (*QudaCommsMap)(const int *coords, void *fdata)

    #
    # @param mycomm User provided MPI communicator in place of MPI_COMM_WORLD
    #

    void qudaSetCommHandle(void *mycomm)

    #
    # Declare the grid mapping ("logical topology" in QMP parlance)
    # used for communications in a multi-GPU grid.  This function
    # should be called prior to initQuda().  The only case in which
    # it's optional is when QMP is used for communication and the
    # logical topology has already been declared by the application.
    #
    # @param nDim   Number of grid dimensions.  "4" is the only supported
    #               value currently.
    #
    # @param dims   Array of grid dimensions.  dims[0]*dims[1]*dims[2]*dims[3]
    #               must equal the total number of MPI ranks or QMP nodes.
    #
    # @param func   Pointer to a user-supplied function that maps coordinates
    #               in the communication grid to MPI ranks (or QMP node IDs).
    #               If the pointer is NULL, the default mapping depends on
    #               whether QMP or MPI is being used for communication.  With
    #               QMP, the existing logical topology is used if it's been
    #               declared.  With MPI or as a fallback with QMP, the default
    #               ordering is lexicographical with the fourth ("t") index
    #               varying fastest.
    #
    # @param fdata  Pointer to any data required by "func" (may be NULL)
    #
    # @see QudaCommsMap
    #

    void initCommsGridQuda(int nDim, const int *dims, QudaCommsMap func, void *fdata)

    #
    # Initialize the library.  This is a low-level interface that is
    # called by initQuda.  Calling initQudaDevice requires that the
    # user also call initQudaMemory before using QUDA.
    #
    # @param device CUDA device number to use.  In a multi-GPU build,
    #               this parameter may either be set explicitly on a
    #               per-process basis or set to -1 to enable a default
    #               allocation of devices to processes.
    #
    void initQudaDevice(int device)

    #
    # Initialize the library persistant memory allocations (both host
    # and device).  This is a low-level interface that is called by
    # initQuda.  Calling initQudaMemory requires that the user has
    # previously called initQudaDevice.
    #
    void initQudaMemory()

    #
    # Initialize the library.  This function is actually a wrapper
    # around calls to initQudaDevice() and initQudaMemory().
    #
    # @param device  CUDA device number to use.  In a multi-GPU build,
    #                this parameter may either be set explicitly on a
    #                per-process basis or set to -1 to enable a default
    #                allocation of devices to processes.
    #
    void initQuda(int device)

    #
    # Finalize the library.
    #
    void endQuda()

    #
    # @brief update the radius for halos.
    # @details This should only be needed for automated testing when
    # different partitioning is applied within a single run.
    #
    void updateR()

    #
    # A new QudaGaugeParam should always be initialized immediately
    # after it's defined (and prior to explicitly setting its members)
    # using this function.  Typical usage is as follows:
    #
    #   QudaGaugeParam gauge_param = newQudaGaugeParam()
    #
    QudaGaugeParam newQudaGaugeParam()

    #
    # A new QudaInvertParam should always be initialized immediately
    # after it's defined (and prior to explicitly setting its members)
    # using this function.  Typical usage is as follows:
    #
    #   QudaInvertParam invert_param = newQudaInvertParam()
    #
    QudaInvertParam newQudaInvertParam()

    #
    # A new QudaMultigridParam should always be initialized immediately
    # after it's defined (and prior to explicitly setting its members)
    # using this function.  Typical usage is as follows:
    #
    #   QudaMultigridParam mg_param = newQudaMultigridParam()
    #
    QudaMultigridParam newQudaMultigridParam()

    #
    # A new QudaEigParam should always be initialized immediately
    # after it's defined (and prior to explicitly setting its members)
    # using this function.  Typical usage is as follows:
    #
    #   QudaEigParam eig_param = newQudaEigParam()
    #
    QudaEigParam newQudaEigParam()

    #
    # A new QudaGaugeObservableParam should always be initialized
    # immediately after it's defined (and prior to explicitly setting
    # its members) using this function.  Typical usage is as follows:
    #
    #   QudaGaugeObservalbeParam obs_param = newQudaGaugeObservableParam();
    #
    QudaGaugeObservableParam newQudaGaugeObservableParam()

    #
    # A new QudaGaugeSmearParam should always be initialized
    # immediately after it's defined (and prior to explicitly setting
    # its members) using this function.  Typical usage is as follows:
    #
    #   QudaGaugeSmearParam smear_param = newQudaGaugeSmearParam();
    #
    QudaGaugeSmearParam newQudaGaugeSmearParam()

    #
    # A new QudaBLASParam should always be initialized immediately
    # after it's defined (and prior to explicitly setting its members)
    # using this function.  Typical usage is as follows:
    #
    #   QudaBLASParam blas_param = newQudaBLASParam()
    #
    QudaBLASParam newQudaBLASParam()

    #
    # Print the members of QudaGaugeParam.
    # @param param The QudaGaugeParam whose elements we are to print.
    #
    void printQudaGaugeParam(QudaGaugeParam *param)

    #
    # Print the members of QudaInvertParam.
    # @param param The QudaInvertParam whose elements we are to print.
    #
    void printQudaInvertParam(QudaInvertParam *param)

    #
    # Print the members of QudaMultigridParam.
    # @param param The QudaMultigridParam whose elements we are to print.
    #
    void printQudaMultigridParam(QudaMultigridParam *param)

    #
    # Print the members of QudaEigParam.
    # @param param The QudaEigParam whose elements we are to print.
    #
    void printQudaEigParam(QudaEigParam *param)

    #
    # Print the members of QudaGaugeObservableParam.
    # @param param The QudaGaugeObservableParam whose elements we are to print.
    #
    void printQudaGaugeObservableParam(QudaGaugeObservableParam *param)

    #
    # Print the members of QudaBLASParam.
    # @param param The QudaBLASParam whose elements we are to print.
    #
    void printQudaBLASParam(QudaBLASParam *param)

    #
    # Load the gauge field from the host.
    # @param h_gauge Base pointer to host gauge field (regardless of dimensionality)
    # @param param   Contains all metadata regarding host and device storage
    #
    void loadGaugeQuda(void *h_gauge, QudaGaugeParam *param)

    #
    # Free QUDA's internal copy of the gauge field.
    #
    void freeGaugeQuda()

    #
    # Free a unique type (Wilson, HISQ fat, HISQ long, smeared) of internal gauge field.
    # @param link_type[in] Type of link type to free up
    #
    void freeUniqueGaugeQuda(QudaLinkType link_type)

    #
    # Free QUDA's internal smeared gauge field.
    #
    void freeGaugeSmearedQuda()

    #
    # Free QUDA's internal two-link gauge field.
    #
    void freeGaugeTwoLinkQuda()

    #
    # Save the gauge field to the host.
    # @param h_gauge Base pointer to host gauge field (regardless of dimensionality)
    # @param param   Contains all metadata regarding host and device storage
    #
    void saveGaugeQuda(void *h_gauge, QudaGaugeParam *param)

    #
    # Load the clover term and/or the clover inverse from the host.
    # Either h_clover or h_clovinv may be set to NULL.
    # @param h_clover    Base pointer to host clover field
    # @param h_cloverinv Base pointer to host clover inverse field
    # @param inv_param   Contains all metadata regarding host and device storage
    #
    void loadCloverQuda(void *h_clover, void *h_clovinv,
                        QudaInvertParam *inv_param)

    #
    # Free QUDA's internal copy of the clover term and/or clover inverse.
    #
    void freeCloverQuda()

    #
    # Perform the solve, according to the parameters set in param.  It
    # is assumed that the gauge field has already been loaded via
    # loadGaugeQuda().
    # @param h_x    Solution spinor field
    # @param h_b    Source spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage and solver parameters
    #
    void lanczosQuda(int k0, int m, void *hp_Apsi, void *hp_r, void *hp_V, void *hp_alpha, void *hp_beta,
                     QudaEigParam *eig_param)

    #
    # Perform the eigensolve. The problem matrix is defined by the invert param, the
    # mode of solution is specified by the eig param. It is assumed that the gauge
    # field has already been loaded via  loadGaugeQuda().
    # @param h_evecs  Array of pointers to application eigenvectors
    # @param h_evals  Host side eigenvalues
    # @param param Contains all metadata regarding the type of solve.
    #
    void eigensolveQuda(void **h_evecs, double_complex *h_evals, QudaEigParam *param)

    #
    # Perform the solve, according to the parameters set in param.  It
    # is assumed that the gauge field has already been loaded via
    # loadGaugeQuda().
    # @param h_x    Solution spinor field
    # @param h_b    Source spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage and solver parameters
    #
    void invertQuda(void *h_x, void *h_b, QudaInvertParam *param)

    #
    # @brief Perform the solve like @invertQuda but for multiple rhs by spliting the comm grid into
    # sub-partitions: each sub-partition invert one or more rhs'.
    # The QudaInvertParam object specifies how the solve should be performed on each sub-partition.
    # Unlike @invertQuda, the interface also takes the host side gauge as input. The gauge pointer and
    # gauge_param are used if for inv_param split_grid[0] * split_grid[1] * split_grid[2] * split_grid[3]
    # is larger than 1, in which case gauge field is not required to be loaded beforehand; otherwise
    # this interface would just work as @invertQuda, which requires gauge field to be loaded beforehand,
    # and the gauge field pointer and gauge_param are not used.
    # @param _hp_x       Array of solution spinor fields
    # @param _hp_b       Array of source spinor fields
    # @param param       Contains all metadata regarding host and device storage and solver parameters
    #
    void invertMultiSrcQuda(void **_hp_x, void **_hp_b, QudaInvertParam *param)

    #
    # Solve for multiple shifts (e.g., masses).
    # @param _hp_x    Array of solution spinor fields
    # @param _hp_b    Source spinor fields
    # @param param  Contains all metadata regarding host and device
    #               storage and solver parameters
    #
    void invertMultiShiftQuda(void **_hp_x, void *_hp_b, QudaInvertParam *param)

    #
    # Setup the multigrid solver, according to the parameters set in param.  It
    # is assumed that the gauge field has already been loaded via
    # loadGaugeQuda().
    # @param param  Contains all metadata regarding host and device
    #               storage and solver parameters
    #
    void* newMultigridQuda(QudaMultigridParam *param)

    #
    # @brief Free resources allocated by the multigrid solver
    # @param mg_instance Pointer to instance of multigrid_solver
    # @param param Contains all metadata regarding host and device
    # storage and solver parameters
    #
    void destroyMultigridQuda(void *mg_instance)

    #
    # @brief Updates the multigrid preconditioner for the new gauge / clover field
    # @param mg_instance Pointer to instance of multigrid_solver
    # @param param Contains all metadata regarding host and device
    # storage and solver parameters, of note contains a flag specifying whether
    # to do a full update or a thin update.
    #
    void updateMultigridQuda(void *mg_instance, QudaMultigridParam *param)

    #
    # @brief Dump the null-space vectors to disk
    # @param[in] mg_instance Pointer to the instance of multigrid_solver
    # @param[in] param Contains all metadata regarding host and device
    # storage and solver parameters (QudaMultigridParam::vec_outfile
    # sets the output filename prefix).
    #
    void dumpMultigridQuda(void *mg_instance, QudaMultigridParam *param)

    #
    # Apply the Dslash operator (D_{eo} or D_{oe}).
    # @param[out] h_out  Result spinor field
    # @param[in] h_in   Input spinor field
    # @param[in] param  Contains all metadata regarding host and device
    #               storage
    # @param[in] parity The destination parity of the field
    #
    void dslashQuda(void *h_out, void *h_in, QudaInvertParam *inv_param, QudaParity parity)

    #
    # Apply the covariant derivative.
    # @param[out] h_out  Result spinor field
    # @param[in] h_in   Input spinor field
    # @param[in] dir    Direction of application
    # @param[in] param  Metadata for host and device storage
    #
    void covDevQuda(void *h_out, void *h_in, int dir, QudaInvertParam *param)

    #
    # Apply the covariant derivative.
    # @param[out] h_out  Result spinor field
    # @param[in] h_in   Input spinor field
    # @param[in] dir    Direction of application
    # @param[in] sym    Apply forward=2, backward=2 or symmetric=3 shift
    # @param[in] param  Metadata for host and device storage
    #
    void shiftQuda(void *h_out, void *h_in, int dir, int sym, QudaInvertParam *param)

    #
    # Apply the spin-taste operator.
    # @param[out] h_out  Result spinor field
    # @param[in] h_in   Input spinor field
    # @param[in] spin   Spin gamma structure
    # @param[in] taste  Taste gamma structure
    # @param[in] param  Metadata for host and device storage
    #
    void spinTasteQuda(void *h_out, void *h_in, int spin, int taste, QudaInvertParam *param)

    #
    # @brief Perform the solve like @dslashQuda but for multiple rhs by spliting the comm grid into
    # sub-partitions: each sub-partition does one or more rhs'.
    # The QudaInvertParam object specifies how the solve should be performed on each sub-partition.
    # Unlike @invertQuda, the interface also takes the host side gauge as
    # input - gauge field is not required to be loaded beforehand.
    # @param _hp_x       Array of solution spinor fields
    # @param _hp_b       Array of source spinor fields
    # @param param       Contains all metadata regarding host and device storage and solver parameters
    # @param parity      Parity to apply dslash on
    #
    void dslashMultiSrcQuda(void **_hp_x, void **_hp_b, QudaInvertParam *param, QudaParity parity)

    #
    # Apply the clover operator or its inverse.
    # @param h_out  Result spinor field
    # @param h_in   Input spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage
    # @param parity The source and destination parity of the field
    # @param inverse Whether to apply the inverse of the clover term
    #
    void cloverQuda(void *h_out, void *h_in, QudaInvertParam *inv_param, QudaParity parity, int inverse)

    #
    # Apply the full Dslash matrix, possibly even/odd preconditioned.
    # @param h_out  Result spinor field
    # @param h_in   Input spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage
    #
    void MatQuda(void *h_out, void *h_in, QudaInvertParam *inv_param)

    #
    # Apply M^{\dag}M, possibly even/odd preconditioned.
    # @param h_out  Result spinor field
    # @param h_in   Input spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage
    #
    void MatDagMatQuda(void *h_out, void *h_in, QudaInvertParam *inv_param)


    #
    # The following routines are temporary additions used by the HISQ
    # link-fattening code.
    #

    void set_dim(int *)
    void pack_ghost(void **cpuLink, void **cpuGhost, int nFace,
                    QudaPrecision precision)

    void computeKSLinkQuda(void* fatlink, void* longlink, void* ulink, void* inlink,
                           double *path_coeff, QudaGaugeParam *param)

    #
    # Compute two-link field
    #
    # @param[out] twolink computed two-link field
    # @param[in] inlink  the external field
    # @param[in] param  Contains all metadata regarding host and device
    #               storage
    #
    void computeTwoLinkQuda(void *twolink, void *inlink, QudaGaugeParam *param)

    #
    # Either downloads and sets the resident momentum field, or uploads
    # and returns the resident momentum field
    #
    # @param[in,out] mom The external momentum field
    # @param[in] param The parameters of the external field
    #
    void momResidentQuda(void *mom, QudaGaugeParam *param)

    #
    # Compute the gauge force and update the momentum field
    #
    # @param[in,out] mom The momentum field to be updated
    # @param[in] sitelink The gauge field from which we compute the force
    # @param[in] input_path_buf[dim][num_paths][path_length]
    # @param[in] path_length One less that the number of links in a loop (e.g., 3 for a staple)
    # @param[in] loop_coeff Coefficients of the different loops in the Symanzik action
    # @param[in] num_paths How many contributions from path_length different "staples"
    # @param[in] max_length The maximum number of non-zero of links in any path in the action
    # @param[in] dt The integration step size (for MILC this is dt*beta/3)
    # @param[in] param The parameters of the external fields and the computation settings
    #
    int computeGaugeForceQuda(void *mom, void *sitelink, int ***input_path_buf, int *path_length, double *loop_coeff,
                              int num_paths, int max_length, double dt, QudaGaugeParam *qudaGaugeParam)

    #
    # Compute the product of gauge links along a path and add to/overwrite the output field
    #
    # @param[in,out] out The output field to be updated
    # @param[in] sitelink The gauge field from which we compute the products of gauge links
    # @param[in] input_path_buf[dim][num_paths][path_length]
    # @param[in] path_length One less that the number of links in a loop (e.g., 3 for a staple)
    # @param[in] loop_coeff Coefficients of the different loops in the Symanzik action
    # @param[in] num_paths How many contributions from path_length different "staples"
    # @param[in] max_length The maximum number of non-zero of links in any path in the action
    # @param[in] dt The integration step size (for MILC this is dt*beta/3)
    # @param[in] param The parameters of the external fields and the computation settings
    #
    int computeGaugePathQuda(void *out, void *sitelink, int ***input_path_buf, int *path_length, double *loop_coeff,
                             int num_paths, int max_length, double dt, QudaGaugeParam *qudaGaugeParam)

    #
    # Compute the traces of products of gauge links along paths using the resident field
    #
    # @param[in,out] traces The computed traces
    # @param[in] sitelink The gauge field from which we compute the products of gauge links
    # @param[in] path_length The number of links in each loop
    # @param[in] loop_coeff Multiplicative coefficients for each loop
    # @param[in] num_paths Total number of loops
    # @param[in] max_length The maximum number of non-zero of links in any path in the action
    # @param[in] factor An overall normalization factor
    #
    void computeGaugeLoopTraceQuda(double_complex *traces, int **input_path_buf, int *path_length, double *loop_coeff,
                                   int num_paths, int max_length, double factor)

    #
    # Evolve the gauge field by step size dt, using the momentum field
    # I.e., Evalulate U(t+dt) = e(dt pi) U(t)
    #
    # @param gauge The gauge field to be updated
    # @param momentum The momentum field
    # @param dt The integration step size step
    # @param conj_mom Whether to conjugate the momentum matrix
    # @param exact Whether to use an exact exponential or Taylor expand
    # @param param The parameters of the external fields and the computation settings
    #
    void updateGaugeFieldQuda(void* gauge, void* momentum, double dt,
                              int conj_mom, int exact, QudaGaugeParam* param)

    #
    # Apply the staggered phase factors to the gauge field.  If the
    # imaginary chemical potential is non-zero then the phase factor
    # exp(imu/T) will be applied to the links in the temporal
    # direction.
    #
    # @param gauge_h The gauge field
    # @param param The parameters of the gauge field
    #
    void staggeredPhaseQuda(void *gauge_h, QudaGaugeParam *param)

    #
    # Project the input field on the SU(3) group.  If the target
    # tolerance is not met, this routine will give a runtime error.
    #
    # @param gauge_h The gauge field to be updated
    # @param tol The tolerance to which we iterate
    # @param param The parameters of the gauge field
    #
    void projectSU3Quda(void *gauge_h, double tol, QudaGaugeParam *param)

    #
    # Evaluate the momentum contribution to the Hybrid Monte Carlo
    # action.
    #
    # @param momentum The momentum field
    # @param param The parameters of the external fields and the computation settings
    # @return momentum action
    #
    double momActionQuda(void* momentum, QudaGaugeParam* param)

    #
    # Allocate a gauge (matrix) field on the device and optionally download a host gauge field.
    #
    # @param gauge The host gauge field (optional - if set to 0 then the gauge field zeroed)
    # @param geometry The geometry of the matrix field to create (1 - scalar, 4 - vector, 6 - tensor)
    # @param param The parameters of the external field and the field to be created
    # @return Pointer to the gauge field (cast as a void*)
    #
    void* createGaugeFieldQuda(void* gauge, int geometry, QudaGaugeParam* param)

    #
    # Copy the QUDA gauge (matrix) field on the device to the CPU
    #
    # @param outGauge Pointer to the host gauge field
    # @param inGauge Pointer to the device gauge field (QUDA device field)
    # @param param The parameters of the host and device fields
    #
    void  saveGaugeFieldQuda(void* outGauge, void* inGauge, QudaGaugeParam* param)

    #
    # Reinterpret gauge as a pointer to a GaugeField and call destructor.
    #
    # @param gauge Gauge field to be freed
    #
    void destroyGaugeFieldQuda(void* gauge)

    #
    # Compute the clover field and its inverse from the resident gauge field.
    #
    # @param param The parameters of the clover field to create
    #
    void createCloverQuda(QudaInvertParam* param)

    #
    # Compute the clover force contributions from a set of partial
    # fractions stemming from a rational approximation suitable for use
    # within MILC.
    #
    # @param mom Force matrix
    # @param dt Integrating step size
    # @param x Array of solution vectors
    # @param p Array of intermediate vectors
    # @param coeff Array of residues for each contribution (multiplied by stepsize)
    # @param kappa2 -kappa*kappa parameter
    # @param ck -clover_coefficient * kappa / 8
    # @param nvec Number of vectors
    # @param multiplicity Number fermions this bilinear reresents
    # @param gauge Gauge Field
    # @param gauge_param Gauge field meta data
    # @param inv_param Dirac and solver meta data
    #
    void computeCloverForceQuda(void *mom, double dt, void **x, void **p, double *coeff, double kappa2, double ck,
                                int nvector, double multiplicity, void *gauge,
                                QudaGaugeParam *gauge_param, QudaInvertParam *inv_param)

    #
    # Compute the force from a clover or twisted clover determinant or
    # a set of partial fractions stemming from a rational approximation
    # suitable for use from within tmLQCD.
    #
    # @param h_mom Host force matrix
    # @param h_x Array of solution vectors x_i = ( Q^2 + s_i )^{-1} b
    # @param h_x0 Array of source vector necessary to compute the force of a ratio of determinant
    # @param coeff Array of coefficients for the rational approximation or {1.0} for the determinant.
    # @param nvector Number of solution vectors and coefficients
    # @param gauge_param Gauge field meta data
    # @param inv_param Dirac and solver meta data
    # @param detratio if 0 compute the force of a determinant otherwise compute the force from a ratio of determinants
    #
    void computeTMCloverForceQuda(void *h_mom, void **h_x, void **h_x0, double *coeff, int nvector,
                                  QudaGaugeParam *gauge_param, QudaInvertParam *inv_param, int detratio)

    #
    # Compute the naive staggered force.  All fields must be in the same precision.
    #
    # @param mom Momentum field
    # @param dt Integrating step size
    # @param delta Additional scale factor when updating momentum (mom += delta * [force]_TA
    # @param gauge Gauge field (at present only supports resident gauge field)
    # @param x Array of single-parity solution vectors (at present only supports resident solutions)
    # @param gauge_param Gauge field meta data
    # @param invert_param Dirac and solver meta data
    #
    void computeStaggeredForceQuda(void *mom, double dt, double delta, void *gauge, void **x, QudaGaugeParam *gauge_param,
                                   QudaInvertParam *invert_param)

    #
    # Compute the fermion force for the HISQ quark action and integrate the momentum.
    # @param momentum        The momentum field we are integrating
    # @param dt              The stepsize used to integrate the momentum
    # @param level2_coeff    The coefficients for the second level of smearing in the quark action.
    # @param fat7_coeff      The coefficients for the first level of smearing (fat7) in the quark action.
    # @param w_link          Unitarized link variables obtained by applying fat7 smearing and unitarization to the original links.
    # @param v_link          Fat7 link variables.
    # @param u_link          SU(3) think link variables.
    # @param quark           The input fermion field.
    # @param num             The number of quark fields
    # @param num_naik        The number of naik contributions
    # @param coeff           The coefficient multiplying the fermion fields in the outer product
    # @param param.          The field parameters.
    #
    void computeHISQForceQuda(void* momentum,
                              double dt,
                              const double level2_coeff[6],
                              const double fat7_coeff[6],
                              const void* const w_link,
                              const void* const v_link,
                              const void* const u_link,
                              void** quark,
                              int num,
                              int num_naik,
                              double** coeff,
                              QudaGaugeParam* param)

    #
    # @brief Generate Gaussian distributed fields and store in the
    # resident gauge field. We create a Gaussian-distributed su(n)
    # field and exponentiate it, e.g., U = exp(sigma * H), where H is
    # the distributed su(n) field and sigma is the width of the
    # distribution (sigma = 0 results in a free field, and sigma = 1 has
    # maximum disorder).

    # @param seed The seed used for the RNG
    # @param sigma Width of Gaussian distrubution
    #
    void gaussGaugeQuda(unsigned long long seed, double sigma)

    #
    # @brief Generate Gaussian distributed fields and store in the
    # resident momentum field. We create a Gaussian-distributed su(n)
    # field, e.g., sigma * H, where H is the distributed su(n) field
    # and sigma is the width of the distribution (sigma = 0 results
    # in a free field, and sigma = 1 has maximum disorder).
    #
    # @param seed The seed used for the RNG
    # @param sigma Width of Gaussian distrubution
    #
    void gaussMomQuda(unsigned long long seed, double sigma)

    #
    # Computes the total, spatial and temporal plaquette averages of the loaded gauge configuration.
    # @param[out] Array for storing the averages (total, spatial, temporal)
    #
    void plaqQuda(double plaq[3])

    #
    # @brief Computes the trace of the Polyakov loop of the current resident field
    # in a given direction.

    # @param[out] ploop Trace of the Polyakov loop in direction dir
    # @param[in] dir Direction of Polyakov loop
    #
    void polyakovLoopQuda(double ploop[2], int dir)

    #
    # Performs a deep copy from the internal extendedGaugeResident field.
    # @param Pointer to externally allocated GaugeField
    #
    void copyExtendedResidentGaugeQuda(void *resident_gauge)

    #
    # Performs gaussian/Wuppertal smearing on a given spinor using the gauge field
    # gaugeSmeared, if it exist, or gaugePrecise if no smeared field is present.
    # @param h_in   Input spinor field
    # @param h_out  Output spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage and operator which will be applied to the spinor
    # @param n_steps Number of steps to apply.
    # @param coeff  Width of the Gaussian distribution
    # @param smear_type Gaussian/Wuppertal smearing
    #
    void performFermionSmearQuda(void *h_out, void *h_in, QudaInvertParam *param, const int n_steps, const double coeff,
                                const QudaFermionSmearType smear_type)

    #
    # LEGACY
    # Performs Wuppertal smearing on a given spinor using the gauge field
    # gaugeSmeared, if it exist, or gaugePrecise if no smeared field is present.
    # @param h_out  Result spinor field
    # @param h_in   Input spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage and operator which will be applied to the spinor
    # @param n_steps Number of steps to apply.
    # @param alpha  Alpha coefficient for Wuppertal smearing.
    #
    void performWuppertalnStep(void *h_out, void *h_in, QudaInvertParam *param, unsigned int n_steps, double alpha)

    #
    # LEGACY
    # Performs gaussian smearing on a given spinor using the gauge field
    # gaugeSmeared, if it exist, or gaugePrecise if no smeared field is present.
    # @param h_in   Input spinor field
    # @param h_out  Output spinor field
    # @param param  Contains all metadata regarding host and device
    #               storage and operator which will be applied to the spinor
    # @param n_steps Number of steps to apply.
    # @param omega  Width of the Gaussian distribution
    #
    void performGaussianSmearNStep(void *h_out, void *h_in, QudaInvertParam *param, const int n_steps, const double omega)

    #
    # Performs APE, Stout, or Over Imroved STOUT smearing on gaugePrecise and stores it in gaugeSmeared
    # @param[in] smear_param Parameter struct that defines the computation parameters
    # @param[in,out] obs_param Parameter struct that defines which
    # observables we are making and the resulting observables.
    #
    void performGaugeSmearQuda(QudaGaugeSmearParam *smear_param, QudaGaugeObservableParam *obs_param)

    #
    # Performs Wilson Flow on gaugePrecise and stores it in gaugeSmeared
    # @param[in] smear_param Parameter struct that defines the computation parameters
    # @param[in,out] obs_param Parameter struct that defines which
    # observables we are making and the resulting observables.
    #
    void performWFlowQuda(QudaGaugeSmearParam *smear_param, QudaGaugeObservableParam *obs_param)

    #
    # @brief Calculates a variety of gauge-field observables.  If a
    # smeared gauge field is presently loaded (in gaugeSmeared) the
    # observables are computed on this, else the resident gauge field
    # will be used.
    # @param[in,out] param Parameter struct that defines which
    # observables we are making and the resulting observables.
    #
    void gaugeObservablesQuda(QudaGaugeObservableParam *param)

    #
    # Public function to perform color contractions of the host spinors x and y.
    # @param[in] x pointer to host data
    # @param[in] y pointer to host data
    # @param[out] result pointer to the 16 spin projections per lattice site
    # @param[in] cType Which type of contraction (open, degrand-rossi, etc)
    # @param[in] param meta data for construction of ColorSpinorFields.
    # @param[in] X spacetime data for construction of ColorSpinorFields.
    #
    void contractQuda(const void *x, const void *y, void *result, const QudaContractType cType, QudaInvertParam *param,
                      const int *X)

    #
    # @param[in] x pointer to host data array
    # @param[in] y pointer to host data array
    # @param[out] result pointer to the spin*spin projections per lattice slice site
    # @param[in] cType Which type of contraction (open, degrand-rossi, etc)
    # @param[in] param meta data for construction of ColorSpinorFields.
    # @param[in] src_colors color dilution parameter
    # @param[in] X local lattice dimansions
    # @param[in] source_position source position array
    # @param[in] number of momentum modes
    # @param[in] mom_modes momentum modes
    # @param[in] fft_type Fourier phase factor type (cos, sin or exp{ikx})
    #
    void contractFTQuda(void **x, void **y, void **result, const QudaContractType cType, void *cs_param_ptr,
                        const int src_colors, const int *X, const int *const source_position, const int n_mom,
                        const int *const mom_modes, const QudaFFTSymmType *const fft_type)


    #
    # @brief Gauge fixing with overrelaxation with support for single and multi GPU.
    # @param[in,out] gauge, gauge field to be fixed
    # @param[in] gauge_dir, 3 for Coulomb gauge fixing, other for Landau gauge fixing
    # @param[in] Nsteps, maximum number of steps to perform gauge fixing
    # @param[in] verbose_interval, print gauge fixing info when iteration count is a multiple of this
    # @param[in] relax_boost, gauge fixing parameter of the overrelaxation method, most common value is 1.5 or 1.7.
    # @param[in] tolerance, torelance value to stop the method, if this value is zero then the method stops when
    # iteration reachs the maximum number of steps defined by Nsteps
    # @param[in] reunit_interval, reunitarize gauge field when iteration count is a multiple of this
    # @param[in] stopWtheta, 0 for MILC criterion and 1 to use the theta value
    # @param[in] param The parameters of the external fields and the computation settings
    #
    int computeGaugeFixingOVRQuda(void *gauge, const unsigned int gauge_dir, const unsigned int Nsteps,
                                  const unsigned int verbose_interval, const double relax_boost, const double tolerance,
                                  const unsigned int reunit_interval, const unsigned int stopWtheta, QudaGaugeParam *param)

    #
    # @brief Gauge fixing with Steepest descent method with FFTs with support for single GPU only.
    # @param[in,out] gauge, gauge field to be fixed
    # @param[in] gauge_dir, 3 for Coulomb gauge fixing, other for Landau gauge fixing
    # @param[in] Nsteps, maximum number of steps to perform gauge fixing
    # @param[in] verbose_interval, print gauge fixing info when iteration count is a multiple of this
    # @param[in] alpha, gauge fixing parameter of the method, most common value is 0.08
    # @param[in] autotune, 1 to autotune the method, i.e., if the Fg inverts its tendency we decrease the alpha value
    # @param[in] tolerance, torelance value to stop the method, if this value is zero then the method stops when
    # iteration reachs the maximum number of steps defined by Nsteps
    # @param[in] stopWtheta, 0 for MILC criterion and 1 to use the theta value
    # @param[in] param The parameters of the external fields and the computation settings
    #
    int computeGaugeFixingFFTQuda(void *gauge, const unsigned int gauge_dir, const unsigned int Nsteps,
                                  const unsigned int verbose_interval, const double alpha, const unsigned int autotune,
                                  const double tolerance, const unsigned int stopWtheta, QudaGaugeParam *param)

    #
    # @brief Strided Batched GEMM
    # @param[in] arrayA The array containing the A matrix data
    # @param[in] arrayB The array containing the B matrix data
    # @param[in] arrayC The array containing the C matrix data
    # @param[in] native boolean to use either the native or generic version
    # @param[in] param The data defining the problem execution.
    #
    void blasGEMMQuda(void *arrayA, void *arrayB, void *arrayC, QudaBoolean native, QudaBLASParam *param)

    #
    # @brief Strided Batched in-place matrix inversion via LU
    # @param[in] Ainv The array containing the A inverse matrix data
    # @param[in] A The array containing the A matrix data
    # @param[in] use_native Boolean to use either the native or generic version
    # @param[in] param The data defining the problem execution.
    #
    void blasLUInvQuda(void *Ainv, void *A, QudaBoolean use_native, QudaBLASParam *param)

    #
    # @brief Flush the chronological history for the given index
    # @param[in] index Index for which we are flushing
    #
    void flushChronoQuda(int index)


    #
    # Create deflation solver resources.
    #
    #
    void* newDeflationQuda(QudaEigParam *param)

    #
    # Free resources allocated by the deflated solver
    #
    void destroyDeflationQuda(void *df_instance)

    void setMPICommHandleQuda(void *mycomm)

    # Parameter set for quark smearing operations
    ctypedef struct QudaQuarkSmearParam:
        pass

    #
    # Performs two-link Gaussian smearing on a given spinor (for staggered fermions).
    # @param[in,out] h_in Input spinor field to smear
    # @param[in] smear_param   Contains all metadata the operator which will be applied to the spinor
    #
    void performTwoLinkGaussianSmearNStep(void *h_in, QudaQuarkSmearParam *smear_param)

    #
    # @brief Performs contractions between a set of quark fields and
    # eigenvectors of the 3-d Laplace operator.
    # @param[in,out] host_sinks An array representing the inner
    # products between the quark fields and the eigen-vector fields.
    # Ordered as [nQuark][nEv][Lt][nSpin][complexity].
    # @param[in] host_quark Array of quark fields we are taking the inner over
    # @param[in] n_quark Number of quark fields
    # @param[in] tile_quark Tile size for quark fields (batch size)
    # @param[in] host_evec Array of eigenvectors we are taking the inner over
    # @param[in] n_evec Number of eigenvectors
    # @param[in] tile_evec Tile size for eigenvectors (batch size)
    # @param[in] inv_param Meta-data structure
    # @param[in] X Lattice dimensions
    #
    void laphSinkProject(double_complex *host_sinks, void **host_quark, int n_quark, int tile_quark,
                        void **host_evec, int nevec, int tile_evec, QudaInvertParam *inv_param, const int X[4])






from typing import List, Union

from ..field import LatticeInfo, LatticeGauge
from ..enum_quda import QudaDslashType

from . import general
from .abstract import Multigrid, StaggeredFermionDirac


class StaggeredDirac(StaggeredFermionDirac):
    def __init__(
        self,
        latt_info: LatticeInfo,
        mass: float,
        tol: float,
        maxiter: int,
        tadpole_coeff: float = 1.0,
        multigrid: Union[List[List[int]], Multigrid, None] = None,
    ) -> None:
        kappa = 1 / 2
        self.tadpole_coeff = tadpole_coeff
        super().__init__(latt_info)
        self.newQudaGaugeParam()
        self.newQudaMultigridParam(multigrid, mass, kappa)
        self.newQudaInvertParam(mass, kappa, tol, maxiter)
        self.setPrecision()
        self.setReconstruct()

    def newQudaGaugeParam(self):
        gauge_param = general.newQudaGaugeParam(self.latt_info)
        gauge_param.staggered_phase_applied = 1
        self.gauge_param = gauge_param

    def newQudaMultigridParam(self, multigrid: Union[List[List[int]], Multigrid, None], mass: float, kappa: float):
        if isinstance(multigrid, Multigrid):
            self.multigrid = multigrid
        elif multigrid is not None:
            geo_block_size = multigrid
            mg_param, mg_inv_param = general.newQudaMultigridParam(
                QudaDslashType.QUDA_STAGGERED_DSLASH, mass, kappa, geo_block_size
            )
            self.multigrid = Multigrid(mg_param, mg_inv_param)
        else:
            self.multigrid = Multigrid(None, None)

    def newQudaInvertParam(self, mass: float, kappa: float, tol: float, maxiter: int):
        invert_param = general.newQudaInvertParam(
            QudaDslashType.QUDA_STAGGERED_DSLASH, mass, kappa, tol, maxiter, 0.0, 1.0, self.multigrid.param
        )
        self.invert_param = invert_param

    def loadGauge(self, gauge: LatticeGauge, thin_update_only: bool = False):
        general.loadStaggeredGauge(gauge, self.gauge_param)
        general.loadMultigrid(self.multigrid, self.invert_param, thin_update_only)

    def freeGauge(self):
        general.freeStaggeredGauge()
        general.freeMultigrid(self.multigrid, self.invert_param)



from abc import ABC, abstractmethod
from typing import List, Literal, NamedTuple, Optional

from pyquda_comm.array import arrayRandomNormalComplex
from ..field import (
    LatticeInfo,
    LatticeMom,
    LatticeFermion,
    LatticeStaggeredFermion,
    MultiLatticeFermion,
    MultiLatticeStaggeredFermion,
)
from ..enum_quda import QUDA_MAX_MULTI_SHIFT, QudaDagType, QudaSolutionType, QudaVerbosity
from ..quda import QudaGaugeParam, QudaInvertParam, invertQuda, MatQuda, invertMultiShiftQuda
from ..dirac import getGlobalPrecision
from ..dirac.abstract import Dirac, FermionDirac, StaggeredFermionDirac


class LoopParam(NamedTuple):
    path: List[List[int]]
    coeff: List[float]


class RationalParam(NamedTuple):
    norm_force: float = 0.0
    residue_force: List[float] = [1.0]
    offset_force: List[float] = [0.0]
    norm_sample: float = 0.0
    residue_sample: List[float] = [1.0]
    offset_sample: List[float] = [0.0]
    norm_action: float = 0.0
    residue_action: List[float] = [1.0]
    offset_action: List[float] = [0.0]


class Action(ABC):
    latt_info: LatticeInfo
    dirac: Dirac
    gauge_param: QudaGaugeParam
    loop_param: LoopParam

    def __init__(self, latt_info: LatticeInfo, dirac: Dirac) -> None:
        self.latt_info = latt_info
        self.dirac = dirac
        self.gauge_param = self.dirac.gauge_param

    @abstractmethod
    def action(self) -> float:
        pass

    @abstractmethod
    def force(self, dt: float, mom: Optional[LatticeMom] = None):
        pass


class FermionAction(Action):
    dirac: FermionDirac
    invert_param: QudaInvertParam
    rational_param: RationalParam
    quark: MultiLatticeFermion
    phi: LatticeFermion
    eta: LatticeFermion

    def __init__(self, latt_info: LatticeInfo, dirac: FermionDirac) -> None:
        super().__init__(latt_info, dirac)
        self.invert_param = self.dirac.invert_param

    def setVerbosity(self, verbosity: QudaVerbosity):
        self.dirac.setVerbosity(verbosity)

    @abstractmethod
    def sample(self):
        pass

    def sampleEta(self):
        self.eta.data = arrayRandomNormalComplex(0.0, 2.0**-0.5, self.eta.shape, self.eta.location)

    def invertMultiShift(self, mode: Literal["sample", "action", "force"]):
        if mode == "sample":
            offset, residue, norm = (
                self.rational_param.offset_sample,
                self.rational_param.residue_sample,
                self.rational_param.norm_sample,
            )
            xx, x, b = self.quark, self.phi, self.eta
        elif mode == "action":
            offset, residue, norm = (
                self.rational_param.offset_action,
                self.rational_param.residue_action,
                self.rational_param.norm_action,
            )
            xx, x, b = self.quark, self.eta, self.phi
        elif mode == "force":
            offset, residue, norm = (
                self.rational_param.offset_force,
                self.rational_param.residue_force,
                self.rational_param.norm_force,
            )
            xx, x, b = self.quark, self.eta, self.phi
        assert len(offset) == len(residue)
        num_offset = len(offset)
        if num_offset > 1:
            tol = self.invert_param.tol
            use_invert_sloppy = tol > 2**-15 and (
                min(map(abs, residue)) / max(map(abs, residue)) > 2**-8
            )  # ? tol_0 > epsilon_short and tol_min / tol_max > epsilon_float / epsilon_short
            self.invert_param.num_offset = num_offset
            self.invert_param.offset = offset + [0.0] * (QUDA_MAX_MULTI_SHIFT - num_offset)
            self.invert_param.residue = residue + [0.0] * (QUDA_MAX_MULTI_SHIFT - num_offset)
            self.invert_param.tol_offset = [
                max(tol * abs(residue[0] / residue[i]), 2e-16 * self.latt_info.GLt**0.5) for i in range(num_offset)
            ] + [0.0] * (QUDA_MAX_MULTI_SHIFT - num_offset)
            if use_invert_sloppy:
                self.invert_param.cuda_prec_sloppy = getGlobalPrecision("invert").sloppy
                self.invert_param.cuda_prec_precondition = getGlobalPrecision("invert").precondition
            invertMultiShiftQuda(xx.even_ptrs, b.even_ptr, self.invert_param)
            self.dirac.performance()
            if use_invert_sloppy:
                self.invert_param.cuda_prec_sloppy = getGlobalPrecision("multishift").sloppy
                self.invert_param.cuda_prec_precondition = getGlobalPrecision("multishift").precondition
            if mode == "sample" or mode == "action":
                x.even = norm * b.even
                for i in range(num_offset):
                    x.even += residue[i] * xx[i].even
        else:  # S = -\phi^\dagger (M^\dagger M)^{-1} \phi
            if mode == "sample":
                self.invert_param.dagger = QudaDagType.QUDA_DAG_YES
                MatQuda(x.even_ptr, b.even_ptr, self.invert_param)
                self.invert_param.dagger = QudaDagType.QUDA_DAG_NO
            elif mode == "action":
                self.invert_param.solution_type = QudaSolutionType.QUDA_MATPC_SOLUTION
                self.invert_param.dagger = QudaDagType.QUDA_DAG_YES
                invertQuda(x.even_ptr, b.even_ptr, self.invert_param)
                self.dirac.performance()
                self.invert_param.solution_type = QudaSolutionType.QUDA_MATPCDAG_MATPC_SOLUTION
                self.invert_param.dagger = QudaDagType.QUDA_DAG_NO
            elif mode == "force":
                invertQuda(xx[0].even_ptr, b.even_ptr, self.invert_param)
                self.dirac.performance()


class StaggeredFermionAction(FermionAction):
    dirac: StaggeredFermionDirac
    quark: Optional[MultiLatticeStaggeredFermion]
    phi: LatticeStaggeredFermion
    eta: LatticeStaggeredFermion

    def __init__(self, latt_info: LatticeInfo, dirac: StaggeredFermionDirac) -> None:
        super().__init__(latt_info, dirac)
